import pysam


def get_read_names(infile):
    '''Returns set of read names from input bam file'''
    samfile = pysam.AlignmentFile(infile, 'rb')
    reads = [x.query_name for x in samfile.fetch(until_eof=True)]
    samfile.close()
    return reads


def get_unmapped_reads(infile, outfile):
    '''Writes BAM file of unmapped reads from infile'''
    pysam.view('-b', '-f', '0x4', '-o', outfile, infile, catch_stdout=False)


def get_region(infile, ref_name, start, end, outfile):
    '''Writes BAM file of the given region'''
    region = ref_name + ':' + str(start + 1) + '-' + str(end + 1)
    pysam.view('-b', '-F', '0x4', '-o', outfile, infile, region, catch_stdout=False)




from minos import dnadiff_mapping_based_verifier

def run(options):
    verifier = dnadiff_mapping_based_verifier.DnadiffMappingBasedVerifier(
        options.dnadiff_snps_file,
        options.dnadiff_ref,
        options.dnadiff_query,
        options.vcf_file1,
        options.vcf_file2,
        options.vcf_ref,
        options.outprefix,
        flank_length=options.flank_length,
        merge_length=options.variant_merge_length,
        filter_and_cluster_vcf=not options.no_filter_cluster,
        discard_ref_calls=not options.include_ref_calls,
        allow_flank_mismatches=options.allow_flank_mismatches,
        exclude_regions_bed_file1=options.exclude_bed1,
        exclude_regions_bed_file2=options.exclude_bed2,
        max_soft_clipped=options.max_soft_clipped,
    )
    verifier.run()




import logging
import os
import shutil

from cluster_vcf_records import vcf_file_read

from minos import dependencies, utils, vcf_file_split_deletions

class Error (Exception): pass

class MultiSamplePipeline:
    def __init__(self,
        ref_fasta,
        input_data_tsv,
        output_dir,
        max_alleles_per_cluster=5000,
        min_large_ref_length=50,
        gramtools_max_read_length=0,
        gramtools_kmer_size=5,
        gramtools_build_threads=1,
        nextflow_config_file=None,
        nextflow_work_dir=None,
        force=False,
        no_run=False,
        clean=True,
        testing=False,
        variants_per_split=None,
        alleles_per_split=None,
        total_splits=None,
        nf_ram_cluster_small_vars=2,
        nf_ram_gramtools_build_small=12,
        nf_ram_minos_small_vars=5,
        nf_ram_merge_small_vars=4,
        use_unmapped_reads=False,
    ):
        self.ref_fasta = os.path.abspath(ref_fasta)
        if not os.path.exists(self.ref_fasta):
            raise Error('Reference FASTA file not found: ' + ref_fasta)

        self.input_data_tsv = os.path.abspath(input_data_tsv)
        if not os.path.exists(self.input_data_tsv):
            raise Error('Data TSV file not found: ' + input_data_tsv)

        self.output_dir = os.path.abspath(output_dir)
        self.nextflow_config_file = None if nextflow_config_file is None else os.path.abspath(nextflow_config_file)
        self.max_alleles_per_cluster = max_alleles_per_cluster
        self.min_large_ref_length = min_large_ref_length
        self.gramtools_max_read_length = gramtools_max_read_length
        self.gramtools_kmer_size = gramtools_kmer_size
        self.gramtools_build_threads = gramtools_build_threads

        if nextflow_work_dir is None:
            self.nextflow_work_dir = os.path.join(self.output_dir, 'nextflow.work')
        else:
            self.nextflow_work_dir = os.path.abspath(nextflow_work_dir)

        self.force = force
        self.nextflow_input_tsv = os.path.join(self.output_dir, 'nextflow.input.tsv')
        self.log_file = os.path.join(self.output_dir, 'log.txt')
        self.no_run = no_run
        self.clean = clean
        self.testing = testing
        self.variants_per_split = variants_per_split
        self.alleles_per_split = alleles_per_split
        self.total_splits = total_splits
        self.nf_ram_cluster_small_vars =  nf_ram_cluster_small_vars
        self.nf_ram_gramtools_build_small = nf_ram_gramtools_build_small
        self.nf_ram_minos_small_vars = nf_ram_minos_small_vars
        self.nf_ram_merge_small_vars = nf_ram_merge_small_vars
        self.use_unmapped_reads = use_unmapped_reads



    @classmethod
    def _load_input_data_tsv(cls, infile):
        logging.info('Start reading file ' + infile)
        data = []
        with open(infile) as f:
            for line in f:
                try:
                    vcf_file, *reads_files = line.rstrip().split('\t')
                except:
                    raise Error('Bad line in input TSV file: ' + line.rstrip())

                if not(os.path.exists(vcf_file)):
                    raise Error('VCF file not found: ' + vcf_file)
                for reads_file in reads_files:
                    if not(os.path.exists(reads_file)):
                        raise Error('Reads file not found: ' + reads_file)

                data.append((os.path.abspath(vcf_file), [os.path.abspath(x) for x in reads_files]))

        logging.info('Finish reading file ' + infile + '. Loaded ' + str(len(data)) + ' samples')
        return data


    @classmethod
    def _merge_vcf_files(cls, infiles_list, outfile):
        '''Reimplementation of bcftools merge. Load all files into
        memory, then write output. bcftools opens all files at the same
        time, which doesn't work for lots of files'''
        vcf_file_data_list_per_sample = []
        sample_names = []
        header_lines = []
        common_first_column_data = []
        first_file = True

        for filename in infiles_list:
            new_data = []

            with open(filename) as f_vcf:
                for vcf_line in f_vcf:
                    if vcf_line.startswith('#'):
                        if first_file and vcf_line.startswith('##'):
                            header_lines.append(vcf_line.rstrip())
                        elif vcf_line.startswith('#CHROM'):
                            fields = vcf_line.rstrip().split('\t')
                            assert len(fields) == 10
                            sample_names.append(fields[-1])
                    else:
                        first_columns, last_column = vcf_line.rstrip().rsplit('\t', maxsplit=1)
                        new_data.append(last_column)
                        if first_file:
                            common_first_column_data.append(first_columns)

            vcf_file_data_list_per_sample.append(new_data)
            first_file = False

        with open(outfile, 'w') as f:
            print(*header_lines, sep='\n', file=f)
            print('#CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'INFO', 'FORMAT', *sample_names, sep='\t', file=f)
            for i, common_first_data in enumerate(common_first_column_data):
                sample_data_cols = [vcf_file_data_list_per_sample[j][i] for j in range(len(vcf_file_data_list_per_sample))]
                print(common_first_data, *sample_data_cols, sep='\t', file=f)


    @classmethod
    def _filter_input_file_for_clustering(cls, infile, outfile):
        header_lines, vcf_records = vcf_file_read.vcf_file_to_dict(infile, sort=True, homozygous_only=False, remove_asterisk_alts=True, remove_useless_start_nucleotides=True)
        with open(outfile, 'w') as f:
            print(*header_lines, sep='\n', file=f)
            for ref_name in vcf_records:
                for vcf_record in vcf_records[ref_name]:
                    if 'MISMAPPED_UNPLACEABLE' in vcf_record.FILTER:
                        continue
                    if vcf_record.FORMAT is None or 'GT' not in vcf_record.FORMAT:
                        logging.warning('No GT in vcf record:' + str(vcf_record))
                        continue

                    genotype = vcf_record.FORMAT['GT']
                    genotypes = genotype.split('/')

                    called_alleles = set(genotypes)
                    if called_alleles == {'0'} or '.' in called_alleles:
                        continue

                    genotypes = sorted([int(x) for x in genotypes])

                    if len(called_alleles) == 1:
                        assert 0 not in genotypes
                        vcf_record.set_format_key_value('GT', '1/1')
                        vcf_record.ALT = [vcf_record.ALT[int(genotypes[0]) - 1]]
                    else:
                        assert len(called_alleles) == 2
                        vcf_record.set_format_key_value('GT', '0/1')
                        if 0 in genotypes:
                            vcf_record.set_format_key_value('GT', '0/1')
                            vcf_record.ALT = [vcf_record.ALT[genotypes[1] - 1]]
                        else:
                            vcf_record.set_format_key_value('GT', '1/2')
                            vcf_record.ALT = [vcf_record.ALT[genotypes[0] - 1], vcf_record.ALT[genotypes[1] - 1]]

                    print(vcf_record, file=f)


    @classmethod
    def _nextflow_helper_process_input_vcf_file(cls, infile, out_small_vars, out_big_vars, out_sample_name, min_large_ref_length):
        splitter = vcf_file_split_deletions.VcfFileSplitDeletions(infile, out_small_vars, out_big_vars, min_large_ref_length=min_large_ref_length)
        splitter.run()
        header_lines = vcf_file_read.get_header_lines_from_vcf_file(infile)
        sample_name = vcf_file_read.get_sample_name_from_vcf_header_lines(header_lines)
        assert sample_name is not None
        max_read_length = None
        for line in header_lines:
            if line.startswith('##minos_max_read_length='):
                max_read_length = int(line.rstrip().split('=')[1])

        with open(out_sample_name, "w") as f:
            sample_name = vcf_file_read.get_sample_name_from_vcf_file(infile)
            assert sample_name is not None
            print(sample_name, file=f)

        return max_read_length


    @classmethod
    def _write_nextflow_data_tsv(cls, data, outfile):
        with open(outfile, 'w') as f:
            print('sample_id', 'vcf_file', 'reads_files', sep='\t', file=f)
            for i, (vcf_file, reads_files) in enumerate(data):
                print(i, vcf_file, ' '.join(reads_files), sep='\t', file=f)


    @classmethod
    def _write_nextflow_script(cls, outfile):
        with open(outfile, 'w') as f:
            print(r'''
            params.data_in_tsv = ""
            params.ref_fasta = ""
            params.min_large_ref_length = 0
            params.gramtools_max_read_length = 0
            params.gramtools_kmer_size = 0
            params.gramtools_build_threads = 1
            params.final_outdir = ""
            params.testing = false
            params.cluster_small_vars_ram = 2
            params.gramtools_build_small_vars_ram = 12
            params.minos_small_vars_ram = 5
            params.pre_cluster_small_vars_merge_ram = 50
            params.pre_cluster_small_vars_merge_threads = 10
            params.merge_small_vars_ram = 4
            params.variants_per_split = 0
            params.alleles_per_split = 0
            params.total_splits = 0
            params.max_alleles_per_cluster = 5000
            params.use_unmapped_reads = false


            if (params.testing) {
                params.pre_cluster_small_vars_merge_threads = 2
            }

            if (params.use_unmapped_reads) {
                use_unmapped_reads = "--use_unmapped_reads"
            }
            else {
                use_unmapped_reads = ""
            }

            data_in_tsv = file(params.data_in_tsv).toAbsolutePath()
            ref_fasta = file(params.ref_fasta).toAbsolutePath()
            final_outdir = file(params.final_outdir).toAbsolutePath()

            if (!data_in_tsv.exists()) {
                exit 1, "Input data TSV file not found: ${params.data_in_tsv} -- aborting"
            }

            if (!ref_fasta.exists()) {
                exit 1, "Reference FASTA file not found: ${params.ref_fasta} -- aborting"
            }

            if (params.min_large_ref_length < 1) {
                exit 1, "Must use option --min_large_ref_length -- aborting"
            }

            if (params.gramtools_kmer_size < 1) {
                exit 1, "Must use option --gramtools_kmer_size -- aborting"
            }

            if (!final_outdir.exists()) {
                exit 1, "Output directory not found: ${params.final_outdir} -- aborting"
            }

            split_tsv = Channel.from(data_in_tsv).splitCsv(header: true, sep:'\t')

            process process_input_vcf_file {
                memory '0.5 GB'
                input:
                val tsv_fields from split_tsv

                output:
                file("small_vars.${tsv_fields['sample_id']}.vcf") into process_input_vcf_file_out_small
                set(val(tsv_fields), file("big_vars.${tsv_fields['sample_id']}.vcf")) into merge_small_and_large_vars_in
                set(val(tsv_fields), file("sample_name.${tsv_fields['sample_id']}")) into minos_all_small_vars_tsv_in
                stdout into max_read_lengths

                """
                #!/usr/bin/env python3
                from minos import multi_sample_pipeline
                multi_sample_pipeline.MultiSamplePipeline._filter_input_file_for_clustering(
                    "${tsv_fields.vcf_file}",
                    'filtered.vcf'
                )
                max_read_length = multi_sample_pipeline.MultiSamplePipeline._nextflow_helper_process_input_vcf_file(
                    'filtered.vcf',
                    "small_vars.${tsv_fields['sample_id']}.vcf",
                    "big_vars.${tsv_fields['sample_id']}.vcf",
                    "sample_name.${tsv_fields['sample_id']}",
                    ${params.min_large_ref_length}
                )
                if max_read_length is None:
                    print(0, end='')
                else:
                    print(max_read_length, end='')
                """
            }


            process pre_cluster_small_vars_merge {
                errorStrategy {task.attempt < 3 ? 'retry' : 'terminate'}
                memory {params.testing ? '0.5 GB' : 1.GB * params.pre_cluster_small_vars_merge_ram * task.attempt}
                maxRetries 3
                cpus {params.testing? 2 : params.pre_cluster_small_vars_merge_threads}

                input:
                val(file_list) from process_input_vcf_file_out_small.collect()

                output:
                file('pre_cluster_small_vars_merge.vcf') into pre_cluster_small_vars_merge_out

                """
                #!/usr/bin/env python3
                from cluster_vcf_records import vcf_merge
                import pyfastaq
                ref_seqs = dict()
                pyfastaq.tasks.file_to_dict("${ref_fasta}", ref_seqs)

                file_list = ["${file_list.join('", "')}"]

                vcf_merge.merge_vcf_files(file_list, ref_seqs, "pre_cluster_small_vars_merge.vcf", threads=${params.pre_cluster_small_vars_merge_threads})
                """

            }

            process cluster_small_vars_vcf {
                errorStrategy {task.attempt < 3 ? 'retry' : 'terminate'}
                memory {params.testing ? '0.5 GB' : 1.GB * params.cluster_small_vars_ram * task.attempt}
                maxRetries 3

                input:
                file('pre_cluster_small_vars_merge.vcf') from pre_cluster_small_vars_merge_out

                output:
                file('small_vars_clustered.vcf') into cluster_small_vars_vcf_out

                """
                #!/usr/bin/env python3
                from cluster_vcf_records import vcf_clusterer
                clusterer = vcf_clusterer.VcfClusterer(["pre_cluster_small_vars_merge.vcf"], "${ref_fasta}", "small_vars_clustered.vcf", max_alleles_per_cluster=${params.max_alleles_per_cluster})
                clusterer.run()
                """
            }


            process split_vcfs {
                errorStrategy {task.attempt < 3 ? 'retry' : 'terminate'}
                memory {params.testing ? '0.5 GB' : 1.GB * params.gramtools_build_small_vars_ram * task.attempt}
                maxRetries 3

                input:
                file('small_vars_clustered.vcf') from cluster_small_vars_vcf_out
                val(max_read_length) from max_read_lengths.max()

                output:
                file('small_vars_clustered.vcf') into minos_adju_splits
                file("gramtools.build_dir/split*.vcf") into gramtools_build_small_vars_vcfs mode flatten

                """
                #!/usr/bin/env python3
                import sys, shutil, glob, os
                from minos import gramtools, vcf_chunker
                if ${params.gramtools_max_read_length} == 0:
                    max_read_length = ${max_read_length}
                else:
                    max_read_length = ${params.gramtools_max_read_length}

                if max_read_length == 0:
                    print('Error! max read length could not be inferred from input VCF files. Must use option --gramtools_max_read_length')
                    sys.exit(1)

                total_splits = ${params.total_splits} if ${params.total_splits} > 0 else None
                variants_per_split = ${params.variants_per_split} if ${params.variants_per_split} > 0 else None
                alleles_per_split = ${params.alleles_per_split} if ${params.alleles_per_split} > 0 else None
                print("total_splits", total_splits)
                print("variants_per_split", variants_per_split)
                print("alleles_per_split", alleles_per_split)

                if total_splits is None and variants_per_split is None and alleles_per_split is None:
                    gramtools.run_gramtools_build(
                        "gramtools.build_dir",
                        "small_vars_clustered.vcf",
                        "${params.ref_fasta}",
                        max_read_length,
                        kmer_size=${params.gramtools_kmer_size},
                    )
                else:
                    chunker = vcf_chunker.VcfChunker(
                        "gramtools.build_dir",
                        vcf_infile="small_vars_clustered.vcf",
                        ref_fasta="${params.ref_fasta}",
                        variants_per_split=variants_per_split,
                        alleles_per_split=alleles_per_split,
                        max_read_length=max_read_length,
                        total_splits=total_splits,
                        flank_length=max_read_length,
                        gramtools_kmer_size=${params.gramtools_kmer_size},
                        threads=${params.gramtools_build_threads},
                    )
                    chunker.make_split_vcf_files()
                """
            }


            // Parallelised gramtools build. The build chunks will go in same dir as split vcfs
            // This is required for `minos adjudicate` so that it recognises that split vcf and build have been done already.
            process gramtools_build_chunks{
                errorStrategy {task.attempt < 3 ? 'retry' : 'terminate'}
                memory {params.testing ? '0.5 GB' : 1.GB * params.gramtools_build_small_vars_ram * task.attempt}
                maxRetries 3
                cpus params.gramtools_build_threads

                input:
                    file vcf_split from gramtools_build_small_vars_vcfs

                output:
                    file("gmtools_build_dir") into gramtools_build_small_vars_out
                

                """
                #!/usr/bin/env python3
                import sys,os,shutil
                from minos import gramtools

                sample_id = "${vcf_split}".split(".")[1]
                par_dir = os.path.dirname(os.path.realpath("${vcf_split}"))
                build_dir = os.path.join(par_dir,"split.{}.gramtools_build".format(sample_id))
                os.symlink(par_dir,"./gmtools_build_dir")
                gramtools.run_gramtools_build(build_dir, "${vcf_split}", "${params.ref_fasta}", 150, "${params.gramtools_kmer_size}")
                """
            }


            // Per sample minos adjudicate. Each of this process will run quasimap/infer on each split, and merge the results into one vcf.
            process minos_all_small_vars {
                //errorStrategy {task.attempt < 3 ? 'retry' : 'terminate'}
                memory {params.testing ? '0.5 GB' : 1.GB * params.minos_small_vars_ram * task.attempt}
                maxRetries 3
                echo true

                input:
                file('small_vars_clustered.vcf') from minos_adju_splits
                val build_dir from gramtools_build_small_vars_out.collect()
                set(val(tsv_fields), file("sample_name.${tsv_fields['sample_id']}")) from minos_all_small_vars_tsv_in

                output:
                file("small_vars.minos.${tsv_fields['sample_id']}") into minos_all_small_vars_out

                """
                sample_name=\$(cat sample_name.${tsv_fields['sample_id']})
                minos_outdir=small_vars.minos.${tsv_fields['sample_id']}
                minos adjudicate ${use_unmapped_reads} --sample_name \$sample_name --gramtools_build_dir ${build_dir[0]} --reads ${tsv_fields['reads_files'].replaceAll(/ /, " --reads ")} \$minos_outdir ${ref_fasta} "small_vars_clustered.vcf"
                """
            }

            // This takes the list per-sample vcfs output by minos_all_small_vars
            // and merges them into a multi-samp vcf.
            process merge_small_vars_vcfs {
                memory '1 GB'
                publishDir path: final_outdir, mode: 'move', overwrite: true

                input:
                val(minos_dir_list) from minos_all_small_vars_out.collect()

                output:
                file('combined_calls.vcf')

                """
                #!/usr/bin/env python3
                # Files end with .N (N=0,1,2,3,...) Sort numerically on this N
                import os
                from minos import multi_sample_pipeline

                minos_dir_list = ["${minos_dir_list.join('", "')}"]
                tuple_list = []
                for filename in minos_dir_list:
                    fields = filename.rsplit('.', maxsplit=1)
                    tuple_list.append((int(fields[1]), filename))
                tuple_list.sort()
                filenames = [os.path.join(x[1], 'debug.calls_with_zero_cov_alleles.vcf')  for x in tuple_list]
                multi_sample_pipeline.MultiSamplePipeline._merge_vcf_files(filenames, 'combined_calls.vcf')
                """
            }


''', file=f)


    def _make_output_dir(self):
        if os.path.exists(self.output_dir):
            if self.force:
                shutil.rmtree(self.output_dir)
            else:
                raise Error('Error! Output directory already exists. ' + self.output_dir)
        os.mkdir(self.output_dir)


    def _prepare_nextflow_input_files(self):
        input_data = MultiSamplePipeline._load_input_data_tsv(self.input_data_tsv)
        MultiSamplePipeline._write_nextflow_data_tsv(input_data, self.nextflow_input_tsv)


    def run(self):
        self._make_output_dir()
        fh = logging.FileHandler(self.log_file, mode='w')
        log = logging.getLogger()
        formatter = logging.Formatter('[minos %(asctime)s %(levelname)s] %(message)s', datefmt='%d-%m-%Y %H:%M:%S')
        fh.setFormatter(formatter)
        log.addHandler(fh)
        dependencies.check_and_report_dependencies(programs=['nextflow'])

        self._prepare_nextflow_input_files()
        original_dir = os.getcwd()
        os.chdir(self.output_dir)
        nextflow_script = 'nextflow.run.nf'
        MultiSamplePipeline._write_nextflow_script(nextflow_script)
        logging.info('Prepared nextflow files. cd ' + self.output_dir)

        nextflow = dependencies.find_binary('nextflow')
        nextflow_command = [
            nextflow, 'run',
            '-work-dir', self.nextflow_work_dir,
            '-with-dag', 'nextflow.out.dag.pdf',
            '-with-trace', 'nextflow.out.trace.txt',
        ]

        if self.nextflow_config_file is not None:
            nextflow_command.extend(['-c', self.nextflow_config_file])

        nextflow_command += [
            nextflow_script,
            '--ref_fasta', self.ref_fasta,
            '--data_in_tsv', self.nextflow_input_tsv,
            '--max_alleles_per_cluster', str(self.max_alleles_per_cluster),
            '--min_large_ref_length', str(self.min_large_ref_length),
            '--final_outdir', self.output_dir,
            '--gramtools_max_read_length', str(self.gramtools_max_read_length),
            '--cluster_small_vars_ram', str(self.nf_ram_cluster_small_vars),
            '--gramtools_build_small_vars_ram', str(self.nf_ram_gramtools_build_small),
            '--gramtools_kmer_size', str(self.gramtools_kmer_size),
            '--gramtools_build_threads', str(self.gramtools_build_threads),
            '--minos_small_vars_ram', str(self.nf_ram_minos_small_vars),
            '--merge_small_vars_ram', str(self.nf_ram_merge_small_vars),
        ]

        if self.testing:
            nextflow_command.append('--testing')

        if self.use_unmapped_reads:
            nextflow_command.append('--use_unmapped_reads')

        if self.variants_per_split is not None:
            nextflow_command.append('--variants_per_split ' + str(self.variants_per_split))
        if self.alleles_per_split is not None:
            nextflow_command.append('--alleles_per_split ' + str(self.alleles_per_split))
        elif self.total_splits is not None:
            nextflow_command.append('--total_splits ' + str(self.total_splits))

        nextflow_command = ' '.join(nextflow_command)

        if self.no_run:
            print('Prepared nextflow pipeline. --no_run used, so not running. The nextflow command to run is:')
            print(nextflow_command)
            return
        else:
            logging.info('Start running nextflow: ' + nextflow_command)
            syscall_process = utils.syscall(nextflow_command)
            logging.info('Finish running nextflow. Writing nextflow stdout/stderr to files')
            with open('nextflow.stdout', 'w') as f:
                print(syscall_process.stdout.rstrip(), file=f)
            with open('nextflow.stderr', 'w') as f:
                print(syscall_process.stderr.rstrip(), file=f)

            logging.info('cd ' + original_dir)

        if self.clean:
            logging.info('Delete nextflow work directory ' + self.nextflow_work_dir)
            shutil.rmtree(self.nextflow_work_dir)
            logging.info('Delete .nextflow directory')
            shutil.rmtree('.nextflow')

        logging.info('Rename .nextflow.log -> nextflow.log')
        os.rename('.nextflow.log', 'nextflow.log')
        os.chdir(original_dir)



import copy
import datetime
import json
import logging
import os
import statistics

from cluster_vcf_records import vcf_file_read

from minos import dependencies, genotyper, utils
from minos import __version__ as minos_version

class Error (Exception): pass


def _build_json_file_is_good(json_build_report):
    '''Returns true iff looks like gramtools build_report.json
    says that gramtools build ran successfully'''
    if not os.path.exists(json_build_report):
        return False

    with open(json_build_report) as f:
        build_report = json.load(f)
        try:
            returned_zero = build_report['gramtools_cpp_build']['return_value_is_0']
        except:
            return False

        return returned_zero


def run_gramtools_build(outdir, vcf_file, ref_file, max_read_length, kmer_size=10):
    '''Runs gramtools build. Makes new directory called 'outdir' for
    the output'''
    gramtools_exe = dependencies.find_binary('gramtools')
    build_command = ' '.join([
        gramtools_exe,
        'build',
        '--gram-directory', outdir,
        '--vcf', vcf_file,
        '--reference', ref_file,
        '--max-read-length', str(max_read_length),
        '--all-kmers',
        '--kmer-size', str(kmer_size),
    ])
    logging.info('Running gramtools build: ' + build_command)
    completed_process = utils.syscall(build_command, allow_fail=True)
    logging.info('Finished running gramtools build. Return code: ' + str(completed_process.returncode))
    build_report = os.path.join(outdir, 'build_report.json')
    ran_ok = _build_json_file_is_good(build_report) and completed_process.returncode == 0
    if not ran_ok:
        logging.info('Error running gramtools build. See build report file ' + build_report)
        raise Error('Error running gramtools build: ' + build_command)

    # Older gramtools called the perl generated VCF file perl_generated_vcf.
    # New gramtools calls it perl_generated.vcf.
    # Whichever one doesn't exist, symlink it to the one that does
    cwd = os.getcwd()
    os.chdir(outdir)
    vcf1 = 'perl_generated_vcf'
    vcf2 = 'perl_generated.vcf'
    if os.path.exists(vcf1):
        assert not os.path.exists(vcf2)
        os.symlink(vcf1, vcf2)
    elif os.path.exists(vcf2):
        assert not os.path.exists(vcf1)
        os.symlink(vcf2, vcf1)
    else:
        message = f'Could not find perl generated VCF file in directory {outdir}. Looked for {vcf1}, {vcf2}. Cannot continue'
        logging.error(message)
        raise Error(message)
    os.chdir(cwd)

    logging.info('Build report file looks good from gramtools build: ' + build_report)


def run_gramtools(build_dir, quasimap_dir, vcf_file, ref_file, reads, max_read_length, kmer_size=10, seed=42):
    '''If build_dir does not exist, runs runs gramtools build and quasimap.
    Otherwise, just runs quasimap. quasimap output is in new
    directory called quasimap_dir.
    "reads" can be one filename, or a list of filenames.
    Raises Error if either of the expected json coverage
    files made by quasimap are not found.'''
    gramtools_exe = dependencies.find_binary('gramtools')
    if not os.path.exists(build_dir):
        run_gramtools_build(build_dir, vcf_file, ref_file, max_read_length, kmer_size=kmer_size)

    if type(reads) is not list:
        assert type(reads) is str
        reads = [reads]

    quasimap_command = ' '.join([
        gramtools_exe,
        'quasimap',
        f'--seed {seed}',
        '--gram-directory', build_dir,
        '--output-directory', quasimap_dir,
        ' '.join(['--reads ' + x for x in reads]),
    ])
    logging.info('Running gramtools quasimap: ' + quasimap_command)
    utils.syscall(quasimap_command)
    logging.info('Finished running gramtools quasimap')

    build_report = os.path.join(build_dir, 'build_report.json')
    quasimap_report = os.path.join(quasimap_dir, 'report.json')
    allele_base_counts_file = os.path.join(quasimap_dir, 'allele_base_coverage.json')
    grouped_allele_counts_file = os.path.join(quasimap_dir, 'grouped_allele_counts_coverage.json')
    files_ok = True
    for filename in build_report, quasimap_report, allele_base_counts_file, grouped_allele_counts_file:
        if not(os.path.exists(filename)):
            files_ok = False
            logging.error('gramtools file not found: ' + filename)

    if not files_ok:
        error_message = 'Looks like something went wrong duing gramtools run. At least one output file not present. Cannot continue.'
        logging.error(error_message)
        raise Error(error_message)

    with open(build_report) as f:
        json_build_report = json.load(f)
    with open(quasimap_report) as f:
        json_quasimap_report = json.load(f)

    return json_build_report, json_quasimap_report


def load_gramtools_vcf_and_allele_coverage_files(vcf_file, quasimap_dir):
    '''Loads the perl_generated_vcf file and allele_coverage files.
    Sanity checks that they agree: 1) same number of lines (excluding header
    lines in vcf) and 2) number of alts agree on each line.
    Raises error at the first time somthing wrong is found.
    Returns a list of tuples: (VcfRecord, dict of allele -> coverage)'''
    allele_base_counts_file = os.path.join(quasimap_dir, 'allele_base_coverage.json')
    grouped_allele_counts_file = os.path.join(quasimap_dir, 'grouped_allele_counts_coverage.json')
    all_allele_coverage, allele_groups = load_allele_files(allele_base_counts_file, grouped_allele_counts_file)
    vcf_header, vcf_lines = vcf_file_read.vcf_file_to_list(vcf_file)
    coverages = []

    if len(all_allele_coverage) != len(vcf_lines):
        raise Error('Number of records in VCF (' + str(len(vcf_lines)) + ') does not match number output from gramtools.(' + str(len(all_allele_coverage)) + '). Cannot continue')

    for i, (allele_combi_coverage, allele_per_base_coverage) in enumerate(all_allele_coverage):
        if len(allele_per_base_coverage) != 1 + len(vcf_lines[i].ALT):
            raise Error('Mismatch in number of alleles for this VCF record:\n' + str(vcf_lines[i]) + '\nLine number is ' + str(i+1))

        coverages.append(sum(allele_combi_coverage.values()))

    assert len(coverages) > 0
    # Unlikely to happen edge case on real data is when coverages has length 1.
    # It happens when running test_run in adjudicator_test, with a split VCf.
    # One of the splits only has 1 record.
    if len(coverages) == 1:
        variance = 1.000
    else:
        variance = round(statistics.variance(coverages), 3)

    return round(statistics.mean(coverages), 3), variance, vcf_header, vcf_lines, all_allele_coverage, allele_groups


def update_vcf_record_using_gramtools_allele_depths(vcf_record, allele_combination_cov, allele_per_base_cov, allele_groups_dict, mean_depth, read_error_rate, kmer_size):
    '''allele_depths should be a dict of allele -> coverage.
    The REF allele must also be in the dict.
    So keys of dict must be equal to REF + ALTs sequences.
    This also changes all columns from QUAL onwards.
    Returns a VcfRecord the same as vcf_record, but with all zero
    coverage alleles removed, and GT and COV fixed accordingly'''
    gtyper = genotyper.Genotyper(mean_depth, read_error_rate, allele_combination_cov, allele_per_base_cov, allele_groups_dict)
    gtyper.run()
    genotype_indexes = set()

    if '.' in gtyper.genotype:
        genotype = './.'
    else:
        if 0 in gtyper.genotype:
            genotype_indexes.add(0)
        for i in range(len(vcf_record.ALT)):
            if i + 1 in gtyper.genotype:
                genotype_indexes.add(i+1)

        if len(genotype_indexes) == 1:
            genotype_index = genotype_indexes.pop()
            genotype = str(genotype_index) + '/' + str(genotype_index)
            genotype_indexes.add(genotype_index)
        else:
            genotype = '/'.join([str(x) for x in sorted(list(genotype_indexes))])

    cov_values = [gtyper.singleton_alleles_cov.get(x, 0) for x in range(1 + len(vcf_record.ALT))]
    cov_string = ','.join([str(x) for x in cov_values])
    vcf_record.QUAL = None
    vcf_record.FILTER = set()
    vcf_record.INFO = {'KMER': str(kmer_size)}
    vcf_record.format_keys = ['DP', 'GT', 'COV', 'GT_CONF']
    vcf_record.FORMAT = {
        'DP': str(sum(allele_combination_cov.values())),
        'GT': genotype,
        'COV': cov_string,
        'GT_CONF': str(gtyper.genotype_confidence)
    }

    # Make new record where all zero coverage alleles are removed
    filtered_record = copy.deepcopy(vcf_record)
    if genotype in ['./.', '0/0']:
        return filtered_record

    indexes_to_keep = set([i for i in range(len(cov_values)) if i == 0 or cov_values[i] > 0])
    indexes_to_keep.update(genotype_indexes)
    indexes_to_keep = list(indexes_to_keep)
    indexes_to_keep.sort()
    filtered_record.FORMAT['COV'] = ','.join([str(cov_values[i]) for i in indexes_to_keep])
    assert indexes_to_keep[0] == 0
    filtered_record.ALT = [filtered_record.ALT[i-1] for i in indexes_to_keep[1:]]

    # The indexes of the genotype string 'n/m' are shifted because
    # we probably removed some alleles
    genotype_strings = {vcf_record.REF if i == 0 else vcf_record.ALT[i-1] for i in genotype_indexes}
    new_genotype_indexes = set()
    if 0 in genotype_indexes:
        new_genotype_indexes.add(0)
    for i, genotype_string in enumerate(filtered_record.ALT):
        if genotype_string in genotype_strings:
            new_genotype_indexes.add(i+1)
            if len(genotype_strings) == len(new_genotype_indexes):
                break

    new_genotype_indexes = list(new_genotype_indexes)
    if len(new_genotype_indexes) == 1:
        new_genotype_indexes.append(new_genotype_indexes[0])
    assert len(new_genotype_indexes) == 2
    filtered_record.FORMAT['GT'] = '/'.join([str(x) for x in new_genotype_indexes])
    return filtered_record


def write_vcf_annotated_using_coverage_from_gramtools(mean_depth, vcf_records, all_allele_coverage, allele_groups, read_error_rate, outfile, kmer_size, sample_name='SAMPLE', max_read_length=None, filtered_outfile=None):
    '''mean_depth, vcf_records, all_allele_coverage, allele_groups should be those
    returned by load_gramtools_vcf_and_allele_coverage_files().
    Writes a new VCF that has allele counts for all the ALTs'''
    assert len(vcf_records) == len(all_allele_coverage)

    header_lines = [
        '##fileformat=VCFv4.2',
        '##source=minos, version ' + minos_version,
        '##fileDate=' + str(datetime.date.today()),
        '##FORMAT=<ID=COV,Number=R,Type=Integer,Description="Number of reads on ref and alt alleles">',
        '##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">',
        '##FORMAT=<ID=DP,Number=1,Type=Integer,Description="total kmer depth from gramtools",Source="minos">',
        '##FORMAT=<ID=GT_CONF,Number=1,Type=Float,Description="Genotype confidence. Difference in log likelihood of most likely and next most likely genotype">',
        '##INFO=<ID=KMER,Number=1,Type=Integer,Description="Kmer size at which variant was discovered (kmer-size used by gramtools build)">',
    ]

    if max_read_length is not None:
        header_lines.append('##minos_max_read_length=' + str(max_read_length))

    header_lines.append('\t'.join(['#CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'INFO', 'FORMAT', sample_name]))

    if filtered_outfile is not None:
        f_filter = open(filtered_outfile, 'w')
        print(*header_lines, sep='\n', file=f_filter)

    with open(outfile, 'w') as f:
        print(*header_lines, sep='\n', file=f)

        for i in range(len(vcf_records)):
            logging.debug('Genotyping: ' + str(vcf_records[i]))
            filtered_record = update_vcf_record_using_gramtools_allele_depths(vcf_records[i], all_allele_coverage[i][0], all_allele_coverage[i][1], allele_groups, mean_depth, read_error_rate, kmer_size)
            print(vcf_records[i], file=f)
            if filtered_outfile is not None:
                print(filtered_record, file=f_filter)

    if filtered_outfile is not None:
        f_filter.close()


def load_allele_files(allele_base_counts_file, grouped_allele_counts_file):
    '''Loads the allele base counts and groupeed allele counts files
    made by gramtools qausimap.'''
    with open(allele_base_counts_file) as f:
        json_base_counts_data = json.load(f)
    with open(grouped_allele_counts_file) as f:
        json_allele_counts_data = json.load(f)

    try:
        allele_base_counts = json_base_counts_data['allele_base_counts']
    except:
        raise Error('Error in json file ' + allele_base_counts_file + '. allele_base_counts not found.')

    try:
        site_counts = json_allele_counts_data['grouped_allele_counts']['site_counts']
    except:
        raise Error('Error in json file ' + grouped_allele_counts_file + '. site_counts not found.')


    try:
        allele_groups = json_allele_counts_data['grouped_allele_counts']['allele_groups']
    except:
         raise Error('Error in json file ' + grouped_allele_counts_file + '. allele_groups not found.')

    if len(allele_base_counts) != len(site_counts):
        raise Error('Mismatch between number of records in json files ' + allele_base_counts_file + ' and ' + grouped_allele_counts_file)

    for key, value in allele_groups.items():
        allele_groups[key] = set(value)

    return list(zip(site_counts, allele_base_counts)), allele_groups



import logging

import matplotlib
matplotlib.use('Agg')
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
matplotlib.style.use('ggplot')

from cluster_vcf_records import vcf_file_read

def load_dp_and_gt_conf_data_from_file(infile):
    return pd.read_csv(infile, header=0, sep='\t')


def scatter_plot_gt_conf_vs_dp(data, outfile):
    p = sns.lmplot(
        x="DP",
        y="GT_CONF",
        data=data,
        fit_reg=False,
        legend=False,
        markers=["o"],
        scatter_kws={"s": 5},
    )
    p.savefig(outfile)
    plt.clf()


def scatter_plot_gt_conf_vs_dp_colour_by_tp_fp(data, outfile, tp_or_fp_types):
    colours = {'TP': 'g', 'FP':'m', 'Unknown': 'black'}
    palette = {x: colours[x] for x in tp_or_fp_types}

    p = sns.lmplot(
        x="DP",
        y="GT_CONF",
        data=data,
        fit_reg=False,
        hue='TP_OR_FP',
        legend=True,
        legend_out=True,
        markers=["o"] * len(palette),
        scatter_kws={"s": 5},
        palette=palette
    )
    p._legend.set_title(None)
    p.savefig(outfile)
    plt.clf()


def histogram_of_one_dataframe_column(data, column_name, outfile):
    p = sns.distplot(data[column_name], kde=False)
    p.figure.savefig(outfile)
    plt.clf()


def histogram_of_one_dataframe_column_color_by_tp_fp(data, column_name, outfile):
    tps = data[data['TP_OR_FP'] == 'TP']
    fps = data[data['TP_OR_FP'] == 'FP']
    p = sns.distplot(tps[column_name], label='True positive', color="g", kde=False)
    p = sns.distplot(fps[column_name], label='False positive', color="m", kde=False)
    plt.legend()
    p.figure.savefig(outfile)
    plt.clf()


def minos_vcf_to_plot_data(infile, outfile):
    header, records = vcf_file_read.vcf_file_to_list(infile)
    data = []
    tp_or_fp_types = set()
    output_cols = ['DP', 'GT_CONF']

    for record in records:
        if record.FORMAT is None:
            continue

        check_geno = record.FORMAT.get('MINOS_CHECK_GENOTYPE', None)
        dp = record.FORMAT.get('DP', None)
        gt_conf = record.FORMAT.get('GT_CONF', None)
        if dp is not None and gt_conf is not None:
            to_append = [dp, gt_conf]

            if check_geno is not None:
                if check_geno == '0':
                    tp_or_fp_type = 'FP'
                elif check_geno == '1':
                    tp_or_fp_type = 'TP'
                else:
                    tp_or_fp_type = 'Unknown'

                tp_or_fp_types.add(tp_or_fp_type)
                to_append.append(tp_or_fp_type)

            data.append(to_append)

    if len(data) == 0:
        logging.warning('No DP and GT_CONF data found in VCF file ' + infile + ' therefore no plots will be made')
        return None

    if 'TP' in tp_or_fp_types or 'FP' in tp_or_fp_types:
        output_cols.append('TP_OR_FP')

    with open(outfile, 'w') as f:
        print(*output_cols, sep='\t', file=f)
        for l in data:
            if len(l) < len(output_cols):
                l.append('Unknown')
            print(*l[:len(output_cols)], sep='\t', file=f)

    return tp_or_fp_types


def plots_from_minos_vcf(infile, outprefix):
    data_tsv = outprefix + '.data.tsv'
    tp_or_fp_types = minos_vcf_to_plot_data(infile, data_tsv)
    if tp_or_fp_types is None:
        return

    data = load_dp_and_gt_conf_data_from_file(data_tsv)
    scatter_file = outprefix + '.gt_conf_dp_scatter.pdf'
    dp_hist_file = outprefix + '.dp_hist.pdf'
    gt_conf_hist_file = outprefix + '.gt_conf_hist.pdf'

    if len(tp_or_fp_types) == 0 or tp_or_fp_types == {'Unknown'}:
        scatter_plot_gt_conf_vs_dp(data, scatter_file)
        histogram_of_one_dataframe_column(data, 'DP', dp_hist_file)
        histogram_of_one_dataframe_column(data, 'GT_CONF', gt_conf_hist_file)
    else:
        scatter_plot_gt_conf_vs_dp_colour_by_tp_fp(data, scatter_file, tp_or_fp_types)
        histogram_of_one_dataframe_column_color_by_tp_fp(data, 'DP', dp_hist_file)
        histogram_of_one_dataframe_column_color_by_tp_fp(data, 'GT_CONF', gt_conf_hist_file)




import logging
import os
import math
import glob

import pyfastaq
import pysam
import pandas as pd

from Bio import pairwise2, SeqIO

from cluster_vcf_records import vcf_clusterer, vcf_file_read

from minos import dependencies, dnadiff, plots, utils

class Error (Exception): pass

class EvaluateRecall:
    '''tto write'''
    def __init__(self, truth_vcf_file, truth_vcf_ref, query_vcf_file, query_vcf_ref, outprefix, flank_length=31, merge_length=None, filter_and_cluster_vcf=True, discard_ref_calls=True, allow_flank_mismatches=True, exclude_regions_bed_file=None, max_soft_clipped=3):
        self.truth_vcf_file = os.path.abspath(truth_vcf_file)
        self.truth_vcf_ref = os.path.abspath(truth_vcf_ref)
        self.query_vcf_file = os.path.abspath(query_vcf_file)
        self.query_vcf_ref = os.path.abspath(query_vcf_ref)
        self.sam_file_out = os.path.abspath(outprefix + '.sam')
        self.seqs_out_truth = os.path.abspath(outprefix + '.truth.fa')
        self.filtered_query_vcf = os.path.abspath(outprefix + '.query.filter.vcf')
        self.filtered_truth_vcf = os.path.abspath(outprefix + '.truth.filter.vcf')
        self.clustered_vcf_query = os.path.abspath(outprefix + '.query.filter.cluster.vcf')
        self.clustered_vcf_truth = os.path.abspath(outprefix + '.truth.filter.cluster.vcf')
        self.seqs_out_query = os.path.abspath(outprefix + '.query.fa')
        self.sam_summary = os.path.abspath(outprefix + '.summary.tsv')
        self.stats_out = os.path.abspath(outprefix + '.stats.tsv')
        self.gt_conf_hist_out = os.path.abspath(outprefix + '.gt_conf_hist.tsv')

        self.flank_length = flank_length
        self.merge_length = flank_length if merge_length is None else merge_length
        self.filter_and_cluster_vcf = filter_and_cluster_vcf
        self.discard_ref_calls = discard_ref_calls
        self.allow_flank_mismatches = allow_flank_mismatches

        if self.filter_and_cluster_vcf:
            self.vcf_to_check_truth = self.clustered_vcf_truth
            self.vcf_to_check_query = self.clustered_vcf_query
        else:
            self.vcf_to_check_truth = self.truth_vcf_file
            self.vcf_to_check_query = self.query_vcf_file

        self.exclude_regions = EvaluateRecall._load_exclude_regions_bed_file(exclude_regions_bed_file)
        self.max_soft_clipped = max_soft_clipped
        self.number_ns = 5

    @classmethod
    def _load_exclude_regions_bed_file(cls, infile):
        regions = {}
        if infile is not None:
            with open(infile) as f:
                for line in f:
                    fields = line.rstrip().split('\t')
                    if fields[0] not in regions:
                        regions[fields[0]] = []
                    start = int(fields[1]) - 1
                    end = int(fields[2]) - 1
                    regions[fields[0]].append(pyfastaq.intervals.Interval(start, end))

            for ref_name in regions:
                pyfastaq.intervals.merge_overlapping_in_list(regions[ref_name])

        return regions

    @classmethod
    def _interval_intersects_an_interval_in_list(cls, interval, interval_list):
        # This could be faster by doing something like a binary search.
        # But we're looking for points in intervals, so fiddly to implement.
        # Not expecting a log interval list, so just do a simple check
        # from start to end for now
        i = 0
        while i < len(interval_list) and interval.start > interval_list[i].end:
            i += 1

        return i < len(interval_list) and interval.intersects(interval_list[i])

    @classmethod
    def _filter_vcf_for_clustering(cls, infile, outfile, discard_ref_calls=True):
        header_lines, vcf_records = vcf_file_read.vcf_file_to_dict(infile, sort=True, homozygous_only=False, remove_asterisk_alts=True, remove_useless_start_nucleotides=True)

        with open(outfile, 'w') as f:
            print(*header_lines, sep='\n', file=f)
            for ref_name in vcf_records:
                for vcf_record in vcf_records[ref_name]:
                    if 'MISMAPPED_UNPLACEABLE' in vcf_record.FILTER:
                        continue
                    if vcf_record.FORMAT is None or 'GT' not in vcf_record.FORMAT:
                        logging.warning('No GT in vcf record:' + str(vcf_record))
                        continue
                    if vcf_record.REF in [".", ""]:
                        continue

                    genotype = vcf_record.FORMAT['GT']
                    genotypes = genotype.split('/')
                    called_alleles = set(genotypes)
                    if len(called_alleles) != 1 or (discard_ref_calls and called_alleles == {'0'}) or '.' in called_alleles:
                        continue

                    if len(vcf_record.ALT) > 1:
                        if called_alleles != {'0'}:
                            vcf_record.set_format_key_value('GT', '1/1')
                            try:
                                vcf_record.ALT = [vcf_record.ALT[int(genotypes[0]) - 1]]
                            except:
                                raise Error('BAD VCf line:' + str(vcf_record))
                        else:
                            vcf_record.set_format_key_value('GT', '0/0')
                            vcf_record.ALT = [vcf_record.ALT[0]]
                    if vcf_record.ALT[0] in [".",""]:
                        continue

                    if vcf_record.FORMAT['GT'] == '0':
                        vcf_record.FORMAT['GT'] = '0/0'
                    elif vcf_record.FORMAT['GT'] == '1':
                        vcf_record.FORMAT['GT'] = '1/1'

                    if 'GL' in vcf_record.FORMAT.keys() and 'GT_CONF' not in vcf_record.FORMAT.keys():
                        likelihoods = vcf_record.FORMAT['GL'].split(',')
                        assert(len(likelihoods) > 2)
                        if called_alleles == {'0'}:
                            vcf_record.set_format_key_value('GT_CONF',str(float(likelihoods[0]) - float(likelihoods[1])))
                        else:
                            vcf_record.set_format_key_value('GT_CONF', str(float(likelihoods[int(genotypes[0])]) - float(likelihoods[0])))
                    if 'SupportFraction' in vcf_record.INFO.keys() and 'GT_CONF' not in vcf_record.FORMAT.keys():
                        vcf_record.set_format_key_value('GT_CONF',
                                                        str(float(vcf_record.INFO['SupportFraction'])*100))
                    print(vcf_record, file=f)


    @classmethod
    def _write_vars_plus_flanks_to_fasta(cls, outfile, vcf_records, ref_seqs, flank_length, ref_only=False, number_ns=0):
        '''Given a dict of vcf records made by vcf_file_read.vcf_file_to_dict(),
        and its correcsponding file of reference sequences, writes a new fasta file
        of each ref seq and inferred variant sequence plus flank_length nucleotides added to
        its start and end. Calls each sequence:
            ref_name.start_position.vcf_list_index.allele_number
        where allele_numbers in same order as VCF, with ref seq = allele 0.'''
        prev_ref_name = None
        prev_ref_pos = None
        j = 0
        with open(outfile, 'w') as f:
            for ref_name in sorted(vcf_records):
                for i, vcf_record in enumerate(vcf_records[ref_name]):
                    start_position, alleles = vcf_record.inferred_var_seqs_plus_flanks(ref_seqs[ref_name], flank_length)

                    for allele_index, allele_seq in enumerate(alleles):
                        if not ref_only or allele_index == 0:
                            seq_name = '.'.join([ref_name, str(start_position + 1), str(j), str(i), str(allele_index)])
                            allele_seq = allele_seq.replace('.','')
                            for n in range(number_ns):
                                allele_seq = "N" + allele_seq + "N"
                            print('>' + seq_name, allele_seq, sep='\n', file=f)
                    if prev_ref_name == ref_name and prev_ref_pos == start_position:
                        j += 1
                    else:
                        j = 0
                    prev_ref_name = ref_name
                    prev_ref_pos = start_position



    @classmethod
    def _map_seqs_to_seqs(cls, seqs_file_ref, seqs_file_query, outfile):
        '''Map seqs_file to ref_file using BWA MEM.
        Output is SAM file written to outfile'''
        bwa_binary = dependencies.find_binary('bwa')
        command = ' '.join([
            bwa_binary, 'index',
            seqs_file_ref,
        ])
        utils.syscall(command)

        command = ' '.join([
            bwa_binary, 'aln',
            seqs_file_ref,
            seqs_file_query,
            '>', outfile + ".tmp",
        ])
        utils.syscall(command)

        command = ' '.join([
            bwa_binary, 'samse',
            seqs_file_ref,
            outfile + ".tmp",
            seqs_file_query,
            '>', outfile,
        ])
        utils.syscall(command)
        #os.unlink(outfile + ".tmp")

    @classmethod
    def _check_if_sam_match_is_good(cls, sam_record, flank_length, query_sequence=None, allow_mismatches=True, max_soft_clipped=3):
        logging.debug(f'Checking SAM: {sam_record}')
        if sam_record.is_unmapped:
            return 'Unmapped'

        if not allow_mismatches:
            try:
                nm = sam_record.get_tag('NM')
            except:
                raise Error('No NM tag found in sam record:' + str(sam_record))

            all_mapped = len(sam_record.cigartuples) == 1 and sam_record.cigartuples[0][0] == 0
            if all_mapped and nm == 0:
                logging.debug('SAM record passed no mismatches allowed check')
                return 'Good'
            else:
                logging.debug('SAM record failed no mismatches allowed check')
                return 'Bad_mismatches'

        # don't allow too many soft clipped bases
        if (sam_record.cigartuples[0][0] == 4 and sam_record.cigartuples[0][1] > max_soft_clipped) \
                or (sam_record.cigartuples[-1][0] == 4 and sam_record.cigartuples[-1][1] > max_soft_clipped):
            logging.debug('SAM record failed soft clipping check')
            return 'Bad_soft_clipped'

        if query_sequence is None:
            query_sequence = sam_record.query_sequence
        assert query_sequence is not None

        # if the query is short, which happens when the variant we
        # are checking is too near the start or end of the ref sequence
        if len(query_sequence) < 2 * flank_length + 1:
            # This is an edge case. We don't really know which part
            # of the query seq we're looking for, so guess
            length_diff = 2 * flank_length - len(query_sequence)

            if sam_record.query_alignment_start < 5:
                alt_seq_end = len(query_sequence) - flank_length - 1
                alt_seq_start = min(alt_seq_end, flank_length - length_diff)
            else:
                alt_seq_start = flank_length
                alt_seq_end = max(alt_seq_start, length_diff + len(query_sequence) - flank_length - 1)
        else:
            alt_seq_start = flank_length
            alt_seq_end = len(query_sequence) - flank_length - 1

        aligned_pairs = sam_record.get_aligned_pairs(with_seq=True)
        logging.debug(f'aligned_pairs: {aligned_pairs}')
        wanted_aligned_pairs = []
        current_pos = 0

        i = 0
        while i < len(query_sequence):
            if aligned_pairs[i][0] is None:
                if alt_seq_start - 1 <= current_pos <= alt_seq_end + 1:
                    wanted_aligned_pairs.append(aligned_pairs[i])
            elif current_pos > alt_seq_end:
                break
            else:
                current_pos = aligned_pairs[i][0]
                if alt_seq_start - 1 <= current_pos <= alt_seq_end + 1:
                    wanted_aligned_pairs.append(aligned_pairs[i])

            i += 1

        logging.debug(f'wanted_aligned_pairs: {wanted_aligned_pairs}')
        assert len(wanted_aligned_pairs) > 0

        for pair in wanted_aligned_pairs:
            if None in pair or query_sequence[pair[0]].upper() != pair[2].upper():
                logging.debug('SAM record failed because mismatch in allele sequence plus 1bp either side')
                return 'Bad_allele_mismatch'

        logging.debug('SAM record passed all checks')
        return 'Good'


    @classmethod
    def _index_vcf(cls, vcffile):
        '''Index VCF file'''
        bgzip_binary = dependencies.find_binary('bgzip')
        command = ' '.join([
            bgzip_binary,
            '-c',
            vcffile,
            '>',
            vcffile + ".gz",
        ])
        utils.syscall(command)

        tabix_binary = dependencies.find_binary('tabix')
        command = ' '.join([
            tabix_binary,
            '-p',
            'vcf',
            vcffile + ".gz",
        ])
        utils.syscall(command)

    @classmethod
    def _parse_sam_file_and_vcf(cls, samfile, query_vcf_file, flank_length, allow_mismatches, exclude_regions=None, max_soft_clipped=3, number_ns=0):
        if  exclude_regions is None:
            exclude_regions = {}

        found = []
        match_flag = []
        correct_allele = []
        gt_conf = []
        allele = []

        samfile_handle = pysam.AlignmentFile(samfile, "r")
        sam_previous_record_name = None
        for sam_record in samfile_handle.fetch(until_eof=True):
            if sam_record.query_name == sam_previous_record_name:
                continue
            sam_previous_record_name = sam_record.query_name
            found_conf = False
            found_allele = False

            # see if excluded region in bed file
            ref, start, ref_num, var_num, allele_num = sam_record.query_name.rsplit('.', maxsplit=5)
            start = int(start) + flank_length
            exclude = False
            for ref_name in exclude_regions.keys():
                end = int(start) + 1
                interval = pyfastaq.intervals.Interval(start, end)
                exclude = EvaluateRecall._interval_intersects_an_interval_in_list(interval,
                                                                                  exclude_regions[ref_name])
            if exclude:
                found.append('Exclude')
                gt_conf.append(0)
                allele.append('0')
                continue

            match = EvaluateRecall._check_if_sam_match_is_good(sam_record,
                                                                    flank_length,
                                                                    query_sequence=sam_record.query_sequence,
                                                                    allow_mismatches=allow_mismatches,
                                                                    max_soft_clipped=max_soft_clipped)
            alignment_start = str(sam_record).split("\t")[3]
            match_flag.append(match)
            if match == 'Good':
                logging.debug('SAM record is a good match')
                logging.debug('SAM record reference is %s' %sam_record.reference_name)
                ref_name, expected_start, vcf_pos_index, vcf_record_index, allele_index = sam_record.reference_name.rsplit('.', maxsplit=4)

                vcf_reader = pysam.VariantFile(query_vcf_file)
                vcf_interval_start = int(expected_start) + int(alignment_start) + flank_length - 2 - number_ns
                vcf_interval_end = int(expected_start) + int(alignment_start) + flank_length - number_ns
                logging.debug('Find VCF records matching ref %s in interval [%i,%i]' %(ref_name, vcf_interval_start, vcf_interval_end))
                for i, vcf_record in enumerate(vcf_reader.fetch(ref_name, vcf_interval_start, vcf_interval_end)):
                    if i == int(vcf_pos_index):
                        sample_name = vcf_record.samples.keys()[0]
                        if 'GT' in vcf_record.format.keys() and len(set(vcf_record.samples[sample_name]['GT'])) == 1:
                            if int(allele_index) == int(vcf_record.samples[sample_name]['GT'][0]):
                                found.append('1')
                                allele.append(str(allele_index))
                                correct_allele.append('1')
                                found_allele = True
                                if 'GT_CONF' in vcf_record.format.keys():
                                    gt_conf.append(int(float(vcf_record.samples[sample_name]['GT_CONF'])))
                                    found_conf = True

            if not found_allele:
                found.append('0')
                allele.append('0')
                correct_allele.append('0')
            if not found_conf:
                gt_conf.append(0)
        assert len(found) == len(gt_conf)
        assert len(found) == len(allele)
        assert len(found) == len(match_flag)
        assert len(found) == len(correct_allele)
        return found, gt_conf, allele, match_flag, correct_allele

    @classmethod
    def _parse_sam_files(cls, truth_vcf_file, samfile, query_vcf_file, outfile, flank_length, allow_mismatches=True, exclude_regions=None, max_soft_clipped=3, number_ns=0):
        '''Input is the original dnadiff snps file of sites we are searching for
        and 2 SAM files made by _map_seqs_to_seqs(), which show mappings of snp sites
        from from the dnadiff snps file to the vcf (i.e. searches if VCF contains an record
        with the appropriate sequence.
        Creates a tsv detailing whether the snp difference could be detected and at what
        GT_CONF threshold.
        '''
        header_lines, vcf_records = vcf_file_read.vcf_file_to_dict(truth_vcf_file, sort=True,
                                                                   homozygous_only=False,
                                                                   remove_asterisk_alts=True,
                                                                   remove_useless_start_nucleotides=True)
        id = []
        ref = []
        alt = []
        for ref_name in vcf_records:
            for record in vcf_records[ref_name]:
                id.append(record.POS)
                ref.append(record.REF)
                alt.append(record.ALT[0])
        query_found, query_conf, query_allele, query_match_flag, query_allele_flag = EvaluateRecall._parse_sam_file_and_vcf(samfile, query_vcf_file,
                                                                                       flank_length, allow_mismatches,
                                                                                       exclude_regions, max_soft_clipped, number_ns)
        assert len(id) == len(query_found)
        out_df = pd.DataFrame({'id': id,
                               'ref': ref,
                               'alt': alt,
                               'query_found': query_found,
                               'query_conf': query_conf,
                               'query_allele': query_allele,
                               'query_match_flag': query_match_flag,
                               'query_allele_correct': query_allele_flag})
        out_df.to_csv(outfile, sep='\t')

    @classmethod
    def _gather_stats(cls, tsv_file):
        stats = {x: 0 for x in ['total', 'found_vars', 'missed_vars', 'excluded_vars']}
        gt_conf_hist = {}

        snps = pd.read_csv(tsv_file, sep='\t')
        for index, line in snps.iterrows():
            stats['total'] += 1
            if (line['query_found'] == 'Exclude'):
                stats['excluded_vars'] += 1
            elif (line['query_found'] == 1 or line['query_found'] == '1'):
                stats['found_vars'] += 1
                gt_confs = [i for i in {line[5]} if not math.isnan(i)]
                gt_conf = None
                if len(gt_confs) > 0:
                    gt_conf = max(gt_confs)
                gt_conf_hist[gt_conf] = gt_conf_hist.get(gt_conf, 0) + 1
            else:
                stats['missed_vars'] += 1
        return stats, gt_conf_hist

    def run(self):
        # Cluster together variants in each vcf
        if self.filter_and_cluster_vcf:
            EvaluateRecall._filter_vcf_for_clustering(self.truth_vcf_file, self.filtered_truth_vcf, self.discard_ref_calls)
            EvaluateRecall._filter_vcf_for_clustering(self.query_vcf_file, self.filtered_query_vcf, self.discard_ref_calls)
            if self.discard_ref_calls:
                clusterer_query = vcf_clusterer.VcfClusterer([self.filtered_query_vcf], self.query_vcf_ref, self.clustered_vcf_query, merge_method='simple', max_distance_between_variants=self.merge_length)
                clusterer_truth = vcf_clusterer.VcfClusterer([self.filtered_truth_vcf], self.truth_vcf_ref, self.clustered_vcf_truth, merge_method='simple', max_distance_between_variants=self.merge_length)
            else:
                clusterer_query = vcf_clusterer.VcfClusterer([self.filtered_query_vcf], self.query_vcf_ref, self.clustered_vcf_query, merge_method='gt_aware', max_distance_between_variants=self.merge_length)
                clusterer_truth = vcf_clusterer.VcfClusterer([self.filtered_truth_vcf], self.truth_vcf_ref, self.clustered_vcf_truth, merge_method='gt_aware', max_distance_between_variants=self.merge_length)
            clusterer_query.run()
            clusterer_truth.run()

        vcf_header, vcf_records_truth = vcf_file_read.vcf_file_to_dict(self.vcf_to_check_truth, sort=True, remove_useless_start_nucleotides=True)
        vcf_header, vcf_records_query = vcf_file_read.vcf_file_to_dict(self.vcf_to_check_query, sort=True, remove_useless_start_nucleotides=True)
        sample_from_header = vcf_file_read.get_sample_name_from_vcf_header_lines(vcf_header)
        if sample_from_header is None:
            sample_from_header = 'sample'
        truth_vcf_ref_seqs = {}
        pyfastaq.tasks.file_to_dict(self.truth_vcf_ref, truth_vcf_ref_seqs)
        query_vcf_ref_seqs = {}
        pyfastaq.tasks.file_to_dict(self.query_vcf_ref, query_vcf_ref_seqs)

        EvaluateRecall._write_vars_plus_flanks_to_fasta(self.seqs_out_truth, vcf_records_truth, truth_vcf_ref_seqs, self.flank_length, ref_only=True)
        EvaluateRecall._write_vars_plus_flanks_to_fasta(self.seqs_out_query, vcf_records_query, query_vcf_ref_seqs, self.flank_length, number_ns=self.number_ns)
        EvaluateRecall._map_seqs_to_seqs(self.seqs_out_query, self.seqs_out_truth, self.sam_file_out)
        #for f in glob.glob(self.seqs_out_truth + '*'):
            #os.unlink(f)
        #for f in glob.glob(self.seqs_out_query + '*'):
            #os.unlink(f)

        EvaluateRecall._index_vcf(self.vcf_to_check_query)
        self.vcf_to_check_query = self.vcf_to_check_query + ".gz"
        EvaluateRecall._parse_sam_files(self.vcf_to_check_truth, self.sam_file_out, self.vcf_to_check_query,
                                        self.sam_summary, self.flank_length,
                                        allow_mismatches=self.allow_flank_mismatches,
                                        exclude_regions=self.exclude_regions,
                                        max_soft_clipped=self.max_soft_clipped,
                                        number_ns=self.number_ns)
        stats, gt_conf_hist = EvaluateRecall._gather_stats(self.sam_summary)
        #os.unlink(self.seqs_out_truth)
        #os.unlink(self.seqs_out_truth)
        #for f in glob.glob(self.vcf_to_check_truth + '*'):
        #    os.unlink(f)
        #for f in glob.glob(self.vcf_to_check_query + '*'):
        #    os.unlink(f)

        # write stats file
        with open(self.stats_out, 'w') as f:
            keys = stats.keys()
            print(*keys, sep='\t', file=f)
            print(*[stats[x] for x in keys], sep='\t', file=f)


        # write GT_CONF histogram files
        with open(self.gt_conf_hist_out, 'w') as f:
            print('GT_CONF\tCount', file=f)
            for gt_conf, count in sorted(gt_conf_hist.items()):
                print(gt_conf, count, sep='\t', file=f)



from pkg_resources import get_distribution

try:
    __version__ = get_distribution('bio-minos').version
except:
    __version__ = 'local'


__all__ = [
    'adjudicator',
    'bam_read_extract',
    'dependencies',
    'genotyper',
    'genotype_confidence_simulator',
    'gramtools',
    'mapping_based_verifier',
    'multi_sample_pipeline',
    'plots',
    'tasks',
    'utils',
    'vcf_chunker',
    'vcf_file_split_deletions',
]

from minos import *



from minos import mapping_based_verifier

def run(options):
    verifier = mapping_based_verifier.MappingBasedVerifier(
        options.vcf_file,
        options.vcf_ref,
        options.truth_ref,
        options.outprefix,
        flank_length=options.flank_length,
        expected_variants_vcf=options.expected_variants_vcf,
        run_dnadiff=options.run_dnadiff,
        filter_and_cluster_vcf=not options.no_filter_cluster,
        discard_ref_calls=not options.include_ref_calls,
        allow_flank_mismatches=not options.no_flank_mismatches,
        merge_length=options.variant_merge_length,
        exclude_regions_bed_file=options.exclude_bed,
        max_soft_clipped=options.max_soft_clipped
    )
    verifier.run()




from minos import multi_sample_pipeline

def run(options):
    pipeline = multi_sample_pipeline.MultiSamplePipeline(
        options.ref_fasta,
        options.data_tsv,
        options.outdir,
        max_alleles_per_cluster=options.max_alleles_per_cluster,
        min_large_ref_length=options.min_large_ref_length,
        gramtools_max_read_length=options.max_read_length,
        gramtools_kmer_size=options.gramtools_kmer_size,
        gramtools_build_threads=options.gramtools_build_threads,
        nextflow_config_file=options.nextflow_config_file,
        nextflow_work_dir=options.nextflow_work_dir,
        force=options.force,
        no_run=options.no_run,
        clean=not options.no_clean,
        variants_per_split=options.variants_per_split,
        alleles_per_split=options.alleles_per_split,
        total_splits=options.total_splits,
        nf_ram_cluster_small_vars=options.nf_ram_cluster_small_vars,
        nf_ram_gramtools_build_small=options.nf_ram_gramtools_build_small,
        nf_ram_minos_small_vars=options.nf_ram_minos_small_vars,
        nf_ram_merge_small_vars=options.nf_ram_merge_small_vars,
        testing=options.testing,
        use_unmapped_reads=options.use_unmapped_reads,
    )
    pipeline.run()



import itertools
import math
import operator

from scipy.stats import poisson


class Error (Exception): pass

class Genotyper:
    def __init__(self, mean_depth, error_rate, allele_combination_cov, allele_per_base_cov, allele_groups_dict):
        self.mean_depth = mean_depth
        self.error_rate = error_rate
        self.allele_combination_cov = allele_combination_cov
        self.allele_per_base_cov = allele_per_base_cov
        self.allele_groups_dict = allele_groups_dict
        self.likelihoods = None
        self.genotype = None
        self.genotype_confidence = None
        self.singleton_alleles_cov = {}


    @classmethod
    def _singleton_alleles_and_coverage(cls, allele_combination_cov, allele_groups_dict):
        singleton_alleles = {}
        for allele_key in allele_combination_cov:
            if len(allele_groups_dict[allele_key]) == 1:
                allele = allele_groups_dict[allele_key].pop()
                allele_groups_dict[allele_key].add(allele)
                singleton_alleles[allele] = allele_combination_cov[allele_key]
        return singleton_alleles


    @classmethod
    def _total_coverage(cls, allele_combination_cov):
        return sum(allele_combination_cov.values())


    @classmethod
    def _coverage_of_one_haploid_allele(cls, allele, allele_combination_cov, allele_groups_dict):
        coverage = 0
        for allele_key in allele_combination_cov:
            if allele in allele_groups_dict[allele_key]:
                coverage += allele_combination_cov[allele_key]
        return coverage


    @classmethod
    def _coverage_of_diploid_alleles(cls, allele1, allele2, allele_combination_cov, allele_groups_dict, singleton_alleles_cov):
        allele1_cov = singleton_alleles_cov.get(allele1, 0)
        allele2_cov = singleton_alleles_cov.get(allele2, 0)
        if allele1_cov + allele2_cov == 0:
            allele1_cov = allele2_cov = 1
        allele1_total_cov = 0
        allele2_total_cov = 0

        for allele_key in allele_combination_cov:
            allele_combination = allele_groups_dict[allele_key]
            if allele_combination.issuperset({allele1, allele2}):
                allele1_total_cov += (allele1_cov / (allele1_cov + allele2_cov) ) * allele_combination_cov[allele_key]
                allele2_total_cov += (allele2_cov / (allele1_cov + allele2_cov) ) * allele_combination_cov[allele_key]
            elif allele1 in allele_combination:
                assert allele2 not in allele_combination
                allele1_total_cov += allele_combination_cov[allele_key]
            elif allele2 in allele_combination:
                assert allele1 not in allele_combination
                allele2_total_cov += allele_combination_cov[allele_key]

        return allele1_total_cov, allele2_total_cov


    @classmethod
    def _log_likelihood_homozygous(cls, mean_depth, allele_depth, total_depth, error_rate, allele_length, non_zeros):
        return sum([
            -mean_depth * (1 + (allele_length - non_zeros) / allele_length),
            allele_depth * math.log(mean_depth),
            -math.lgamma(allele_depth + 1),
            (total_depth - allele_depth) * math.log(error_rate),
            non_zeros * math.log(1 - poisson.pmf(0, mean_depth)) / allele_length,
        ])



    @classmethod
    def _log_likelihood_heterozygous(cls, mean_depth, allele_depth1, allele_depth2, total_depth,
            error_rate, allele_length1, allele_length2, non_zeros1, non_zeros2):
        return sum([
            -mean_depth * (1 + 0.5 * ( (1 - (non_zeros1 / allele_length1) ) + (1 - (non_zeros2 / allele_length2) ))),
            (allele_depth1 + allele_depth2) * math.log(0.5 * mean_depth),
            -math.lgamma(allele_depth1 + 1),
            -math.lgamma(allele_depth2 + 1),
            (total_depth - allele_depth1 - allele_depth2) * math.log(error_rate),
            ( (non_zeros1 / allele_length1) + (non_zeros2 / allele_length2) ) * math.log(1 - poisson.pmf(0, 0.5 * mean_depth)),
        ])


    @classmethod
    def _non_zeros_from_allele_per_base_cov(cls, allele_per_base_cov, mean_depth, error_rate):
        non_zeros = []
        for per_base_cov in allele_per_base_cov:
            non_zero_count = 0
            for cov in per_base_cov:
                if  poisson.pmf(cov, mean_depth) > pow(error_rate, cov):
                    non_zero_count += 1
            non_zeros.append(non_zero_count)
        return non_zeros


    def _calculate_log_likelihoods(self):
        '''Makes a list of tuples: ( (allele(s) tuple), log likelihood).
        List is sorted from most to least likely'''
        self.likelihoods = []
        total_depth = sum(self.allele_combination_cov.values())
        non_zeros_per_allele = Genotyper._non_zeros_from_allele_per_base_cov(self.allele_per_base_cov, self.mean_depth, self.error_rate)

        for allele_number, per_base_cov in enumerate(self.allele_per_base_cov):
            allele_depth = Genotyper._coverage_of_one_haploid_allele(allele_number, self.allele_combination_cov, self.allele_groups_dict)
            allele_length = len(per_base_cov)
            non_zeros = non_zeros_per_allele[allele_number]

            log_likelihood = Genotyper._log_likelihood_homozygous(
                    self.mean_depth,
                    allele_depth,
                    total_depth,
                    self.error_rate,
                    allele_length,
                    non_zeros,
            )
            self.likelihoods.append(({allele_number, allele_number}, log_likelihood))

        self.singleton_alleles_cov = Genotyper._singleton_alleles_and_coverage(self.allele_combination_cov, self.allele_groups_dict)


        for (allele_number1, allele_number2) in itertools.combinations(self.singleton_alleles_cov.keys(), 2):
            allele1_depth, allele2_depth = Genotyper._coverage_of_diploid_alleles(allele_number1, allele_number2, self.allele_combination_cov, self.allele_groups_dict, self.singleton_alleles_cov)
            allele1_length = len(self.allele_per_base_cov[allele_number1])
            allele2_length = len(self.allele_per_base_cov[allele_number2])
            non_zeros1 = non_zeros_per_allele[allele_number1]
            non_zeros2 = non_zeros_per_allele[allele_number2]

            log_likelihood = Genotyper._log_likelihood_heterozygous(
                    self.mean_depth,
                    allele1_depth,
                    allele2_depth,
                    total_depth,
                    self.error_rate,
                    allele1_length,
                    allele2_length,
                    non_zeros1,
                    non_zeros2,
            )
            self.likelihoods.append((set([allele_number1, allele_number2]), log_likelihood))

        self.likelihoods.sort(key=operator.itemgetter(1),reverse=True)


    def run(self):
        if len(self.allele_combination_cov) == 0 or Genotyper._total_coverage(self.allele_combination_cov) == 0:
            self.genotype = {'.'}
            self.genotype_confidence = 0.0
        else:
            self._calculate_log_likelihoods()
            assert self.likelihoods is not None and len(self.likelihoods) > 1
            self.genotype, best_log_likelihood = self.likelihoods[0]

            for allele in self.genotype:
                if self.singleton_alleles_cov.get(allele, 0) == 0:
                    self.genotype = {'.'}
                    self.genotype_confidence = 0.0
                    break
            else:
                self.genotype_confidence = round(best_log_likelihood - self.likelihoods[1][1], 2)




from minos import adjudicator

def run(options):
    adj = adjudicator.Adjudicator(
        options.outdir,
        options.ref_fasta,
        options.reads,
        options.vcf_files,
        max_read_length=options.max_read_length,
        read_error_rate=options.read_error_rate,
        overwrite_outdir=options.force,
        max_alleles_per_cluster=options.max_alleles_per_cluster,
        gramtools_build_dir=options.gramtools_build_dir,
        sample_name=options.sample_name,
        variants_per_split=options.variants_per_split,
        alleles_per_split=options.alleles_per_split,
        total_splits=options.total_splits,
        clean=not options.debug,
        gramtools_kmer_size=options.gramtools_kmer_size,
        use_unmapped_reads=options.use_unmapped_reads,
    )
    adj.run()




import logging
import numpy as np
from scipy import stats

from minos import genotyper


class GenotypeConfidenceSimulator:
    def __init__(self, mean_depth, depth_variance, error_rate, allele_length=1, iterations=10000):
        self.mean_depth = mean_depth
        self.depth_variance = depth_variance
        self.error_rate = error_rate
        self.iterations = iterations
        self.allele_length = allele_length
        self.confidence_scores_percentiles = {}
        self.min_conf_score = None
        self.max_conf_score = None


    @classmethod
    def _simulate_confidence_scores(cls, mean_depth, depth_variance, error_rate, iterations, allele_length=1, seed=42):
        np.random.seed(seed)
        allele_groups_dict = {'1': {0}, '2': {1}}
        i = 0
        confidences = []
        # We can't use the negative binomial unless depth_variance > mean_depth.
        # So force it to be so.
        if depth_variance < mean_depth:
            depth_variance = 2 * mean_depth
            logging.warn('Variance in read depth is smaller than mean read depth. Setting variance = 2 * mean, so that variant simulations can run. GT_CONF_PERCENTILE in the output VCF file may not be very useful as a result of this.')
        no_of_successes = (mean_depth ** 2) / (depth_variance - mean_depth)
        prob_of_success = 1 - (depth_variance - mean_depth) / depth_variance

        while i < iterations:
            correct_coverage = np.random.negative_binomial(no_of_successes, prob_of_success)
            incorrect_coverage = np.random.binomial(mean_depth, error_rate)
            if correct_coverage + incorrect_coverage == 0:
                continue

            allele_combination_cov = {}
            if incorrect_coverage > 0:
                allele_combination_cov['1'] = incorrect_coverage
            if correct_coverage > 0:
                allele_combination_cov['2'] = correct_coverage
            allele_per_base_cov = [[incorrect_coverage] * allele_length, [correct_coverage] * allele_length]
            gtyper = genotyper.Genotyper(mean_depth, error_rate, allele_combination_cov, allele_per_base_cov, allele_groups_dict)
            gtyper.run()
            confidences.append(round(gtyper.genotype_confidence))
            i += 1

        assert len(confidences) == iterations
        confidences.sort()
        return confidences


    @classmethod
    def _make_conf_to_percentile_dict(cls, confidence_scores):
        assert len(confidence_scores) > 0
        confidence_scores_percentiles = 100 * stats.rankdata(confidence_scores) / len(confidence_scores)
        conf_to_percentile = {}
        for conf, percentile in zip(confidence_scores, confidence_scores_percentiles):
            conf_to_percentile[conf] = round(percentile, 2)
        return conf_to_percentile


    def get_percentile(self, confidence):
        percentile = self.confidence_scores_percentiles.get(confidence, None)
        if percentile is not None:
            return percentile

        # If we don't know this percentile, then infer it by linearly interpolating
        # between the two nearest values, and add
        # it to the dict, so we don't have to work it out again
        if self.min_conf_score is None or self.max_conf_score is None:
            self.min_conf_score = min(self.confidence_scores_percentiles.keys())
            self.max_conf_score = max(self.confidence_scores_percentiles.keys())

        if confidence < self.min_conf_score:
            self.confidence_scores_percentiles[confidence] = 0.00
        elif confidence > self.max_conf_score:
            self.confidence_scores_percentiles[confidence] = 100.00
        else:
            i = 0
            left_conf = None
            right_conf = None

            for i in range(1, self.max_conf_score, 1):
                if left_conf is None:
                    left_conf = confidence - i if confidence - i in self.confidence_scores_percentiles else None
                if right_conf is None:
                    right_conf = self.confidence_scores_percentiles.get(confidence + 1, None)
                    right_conf = confidence + i if confidence + i in self.confidence_scores_percentiles else None
                if left_conf is not None and right_conf is not None:
                    break
            else:
                raise Exception(f'Error inferring new confidence score {confidence}. Cannot continue')

            assert self.min_conf_score <=  left_conf < confidence < right_conf <= self.max_conf_score
            left_pc = self.confidence_scores_percentiles[left_conf]
            right_pc = self.confidence_scores_percentiles[right_conf]
            if left_pc == right_pc:
                self.confidence_scores_percentiles[confidence] = left_pc
            else:
                self.confidence_scores_percentiles[confidence] = round(left_pc + ((confidence - left_conf) / (right_conf - left_conf)) * (right_pc - left_pc), 2)

        return self.confidence_scores_percentiles[confidence]


    def run_simulations(self):
        confidence_scores = GenotypeConfidenceSimulator._simulate_confidence_scores(self.mean_depth, self.depth_variance, self.error_rate, self.iterations, allele_length=self.allele_length)
        self.confidence_scores_percentiles = GenotypeConfidenceSimulator._make_conf_to_percentile_dict(confidence_scores)



# This class is here for when we were simulating different allele lengths.
# It was used before the genotyper was changed to normalise non-zero coverage positions
# by allele length. It's no longer used, but leave it here just in case.
# Note that using it is currently pointless because the genotyper normalises by length,
# so would only make sense to use it if the genotyper was changed.
class Simulations:
    def __init__(self, mean_depth, depth_variance, error_rate, allele_lengths=[1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,20,30,40,50], iterations=1000):
        self.sims = {l: GenotypeConfidenceSimulator(mean_depth, depth_variance, error_rate, allele_length=l, iterations=iterations) for l in allele_lengths}

        for allele_length in sorted(self.sims):
            logging.info(f'Run simulation, mean_depth={mean_depth}, depth_variance={depth_variance}, error_rate={error_rate}, allele_length={allele_length}, iterations={iterations}')
            self.sims[allele_length].run_simulations()
        logging.info('Finished running simulations')
        self.max_allele_length = max(self.sims.keys())
        self.nearest_allele_length = {}


    def get_percentile(self, allele_length, confidence):
        if allele_length in self.sims:
            return self.sims[allele_length].get_percentile(confidence)
        elif allele_length > self.max_allele_length:
            return self.sims[self.max_allele_length].get_percentile(confidence)
        elif allele_length in self.nearest_allele_length:
            return self.sims[self.nearest_allele_length[allele_length]].get_percentile(confidence)
        else:
            # Find the nearest allele length to the one we want, because
            # we haven't had this allele length before
            nearest_length = None

            for i in range(1, self.max_allele_length, 1):
                if (allele_length - i) in self.sims:
                    nearest_length = allele_length - i
                    break
                elif (allele_length + i) in self.sims:
                    nearest_length = allele_length + i
                    break

            assert nearest_length is not None

            self.nearest_allele_length[allele_length] = nearest_length
            return self.sims[nearest_length].get_percentile(confidence)





__all__ = [
    'adjudicate',
    'check_with_ref',
    'check_snps',
    'check_recall',
    'cluster_vcfs',
    'make_split_gramtools_build',
    'multi_sample_pipeline',
    'versions',
]

from minos.tasks import *



import logging
import os

import pyfastaq
import pysam

from Bio import pairwise2

from cluster_vcf_records import vcf_clusterer, vcf_file_read

from minos import dependencies, dnadiff, plots, utils

class Error (Exception): pass


def sam_reader(infile):
    '''Reads SAM file made by _MappingBasedVerifier._map_seqs_to_ref().
    yields list of all sam records corresponding to each VCF record.'''
    samfile = pysam.AlignmentFile(infile, "r")
    current_records = []
    current_ref_start_record_index = None

    for sam_record in samfile.fetch(until_eof=True):
        ref_name, expected_start, vcf_record_index, allele_index = sam_record.query_name.rsplit('.', maxsplit=3)
        ref_start_record_index = (ref_name, expected_start, vcf_record_index)

        if current_ref_start_record_index in [None, ref_start_record_index]:
            current_records.append(sam_record)
        else:
            yield current_records
            current_records = [sam_record]

        current_ref_start_record_index = ref_start_record_index

    if len(current_records) > 0:
        yield current_records


class MappingBasedVerifier:
    '''vcf_file_in = input VCF file, to be verified.
    vcf_reference_file = reference sequence file that was used to make vcf_file_in.
    verify_reference_file = reference file that we assume is the truth.
    outprefix = prefix of output files.
    flank_length = length of "truth" sequence to take before/after alleles when mapping.

    Writes 3 files:
    outprefix.sam = mapping of seqs + flanks to verify_reference_file.
    outprefix.vcf = VCF file with annotations for validation
    outprefix.stats.tsv = summary stats (see dict output by
                          _parse_sam_file_and_update_vcf_records_and_gather_stats()
                          for a description)'''
    def __init__(self, vcf_file_in, vcf_reference_file, verify_reference_file, outprefix, flank_length=31, merge_length=None, expected_variants_vcf=None, run_dnadiff=True, filter_and_cluster_vcf=True, discard_ref_calls=True, allow_flank_mismatches=True, exclude_regions_bed_file=None, max_soft_clipped=3):
        self.vcf_file_in = os.path.abspath(vcf_file_in)
        self.vcf_reference_file = os.path.abspath(vcf_reference_file)
        self.verify_reference_file = os.path.abspath(verify_reference_file)
        self.vcf_file_out = os.path.abspath(outprefix + '.vcf')
        self.vcf_file_plots_out = os.path.abspath(outprefix + '.vcf.plots')
        self.sam_file_out = os.path.abspath(outprefix + '.sam')
        self.filtered_vcf = os.path.abspath(outprefix + '.filter.vcf')
        self.clustered_vcf = os.path.abspath(outprefix + '.filter.cluster.vcf')
        self.seqs_out = os.path.abspath(outprefix + '.fa')
        self.stats_out = os.path.abspath(outprefix + '.stats.tsv')
        self.gt_conf_hists_filenames = {
            'TP': os.path.abspath(outprefix + '.gt_conf_hist.TP.tsv'),
            'FP': os.path.abspath(outprefix + '.gt_conf_hist.FP.tsv'),
        }
        self.flank_length = flank_length
        self.merge_length = flank_length if merge_length is None else merge_length
        self.expected_variants_vcf = expected_variants_vcf
        self.run_dnadiff = run_dnadiff
        if self.run_dnadiff and self.expected_variants_vcf is not None:
            raise Error('Error! Incompatible options. expected_variants_vcf file provided, and run_dnadiff is True')
        self.dnadiff_outprefix = os.path.abspath(outprefix + '.dnadiff')
        self.vcf_false_negatives_file_out = os.path.abspath(outprefix + '.false_negatives.vcf')
        self.filter_and_cluster_vcf = filter_and_cluster_vcf
        self.discard_ref_calls = discard_ref_calls
        self.allow_flank_mismatches = allow_flank_mismatches

        if self.filter_and_cluster_vcf:
            self.vcf_to_check = self.clustered_vcf
        else:
            self.vcf_to_check = self.vcf_file_in

        self.exclude_regions = MappingBasedVerifier._load_exclude_regions_bed_file(exclude_regions_bed_file)
        self.max_soft_clipped = max_soft_clipped


    @classmethod
    def _needleman_wunsch(cls, seq1, seq2, match=1, mismatch=-3, gap_open=-11, gap_extend=-4):
        '''Returns global alignment strings from NM alignment of the
        two sequences. Dashes for gaps'''
        alignments = pairwise2.align.globalms(seq1, seq2, match, mismatch, gap_open, gap_extend)
        assert len(alignments[0][0]) == len(alignments[0][1])
        return alignments[0][0], alignments[0][1]


    @classmethod
    def _edit_distance_from_alignment_strings(cls, str1, str2):
        '''Input should be seqs output by _needleman_wunsch().
        Returns the edit distance between the sequences'''
        assert len(str1) == len(str2)
        edit_distance = 0
        in_gap = False

        for i, char1 in enumerate(str1):
            if char1 == '-' or str2[i] == '-':
                if not in_gap:
                    in_gap = True
                    edit_distance += 1
            else:
                in_gap = False

                if char1 != str2[i]:
                    edit_distance += 1

        return edit_distance


    @classmethod
    def _edit_distance_between_seqs(cls, seq1, seq2):
        '''Input is two strings). They are globally aligned
        and the edit distance is returned. An indel of any length
        is counted as one edit'''
        aln1, aln2 = MappingBasedVerifier._needleman_wunsch(seq1, seq2)
        return MappingBasedVerifier._edit_distance_from_alignment_strings(aln1, aln2)


    @classmethod
    def _add_edit_distances_to_vcf_record(cls, record):
        '''Adds EDIT_DIST => list of numbers, one for each allele, to the INFO column.
        Also adds GT_EDIT_DIST => float, which is the edit distance between the genotyped allele
        and the ref. If het call, takes the mean edit distance of the two calls'''
        edit_distances = [MappingBasedVerifier._edit_distance_between_seqs(record.REF, x) for x in record.ALT]
        record.INFO['EDIT_DIST'] = ','.join([str(x) for x in edit_distances])
        if 'GT' in record.FORMAT:
            genotypes = set([int(x) for x in record.FORMAT['GT'].split('/')])
            geno_edit_distances = []
            for genotype in genotypes:
                if genotype == 0:
                    geno_edit_distances.append(0)
                else:
                    geno_edit_distances.append(edit_distances[genotype - 1])
            gt_edit_dist = round(sum(geno_edit_distances) / len(geno_edit_distances), 1)
            record.set_format_key_value('GT_EDIT_DIST', str(gt_edit_dist))


    @classmethod
    def _load_exclude_regions_bed_file(cls, infile):
        regions = {}
        if infile is not None:
            with open(infile) as f:
                for line in f:
                    fields = line.rstrip().split('\t')
                    if fields[0] not in regions:
                        regions[fields[0]] = []
                    start = int(fields[1])
                    end = int(fields[2]) - 1
                    regions[fields[0]].append(pyfastaq.intervals.Interval(start, end))

            for ref_name in regions:
                pyfastaq.intervals.merge_overlapping_in_list(regions[ref_name])

        return regions


    @classmethod
    def _interval_intersects_an_interval_in_list(cls, interval, interval_list):
        # This could be faster by doing something like a binary search.
        # But we're looking for points in intervals, so fiddly to implement.
        # Not expecting a log interval list, so just do a simple check
        # from start to end for now
        i = 0
        while i < len(interval_list) and interval.start > interval_list[i].end:
            i += 1

        return i < len(interval_list) and interval.intersects(interval_list[i])


    @classmethod
    def _filter_vcf_for_clustering(cls, infile, outfile, discard_ref_calls=True):
        header_lines, vcf_records = vcf_file_read.vcf_file_to_dict(infile, sort=True, homozygous_only=False, remove_asterisk_alts=True, remove_useless_start_nucleotides=True)

        with open(outfile, 'w') as f:
            print(*header_lines, sep='\n', file=f)
            for ref_name in vcf_records:
                for vcf_record in vcf_records[ref_name]:
                    if 'MISMAPPED_UNPLACEABLE' in vcf_record.FILTER:
                        continue
                    if vcf_record.FORMAT is None or 'GT' not in vcf_record.FORMAT:
                        logging.warning('No GT in vcf record:' + str(vcf_record))
                        continue
                    if vcf_record.REF in [".", ""]:
                        continue

                    genotype = vcf_record.FORMAT['GT']
                    genotypes = genotype.split('/')
                    called_alleles = set(genotypes)
                    if len(called_alleles) != 1 or (discard_ref_calls and called_alleles == {'0'}) or '.' in called_alleles:
                        continue

                    if len(vcf_record.ALT) > 1:
                        if called_alleles != {'0'}:
                            vcf_record.set_format_key_value('GT', '1/1')
                            try:
                                vcf_record.ALT = [vcf_record.ALT[int(genotypes[0]) - 1]]
                            except:
                                raise Error('BAD VCf line:' + str(vcf_record))
                        else:
                            vcf_record.set_format_key_value('GT', '0/0')
                            vcf_record.ALT = [vcf_record.ALT[0]]
                    if vcf_record.ALT[0] in [".",""]:
                        continue

                    if vcf_record.FORMAT['GT'] == '0':
                        vcf_record.FORMAT['GT'] = '0/0'
                    elif vcf_record.FORMAT['GT'] == '1':
                        vcf_record.FORMAT['GT'] = '1/1'

                    if 'GL' in vcf_record.FORMAT.keys() and 'GT_CONF' not in vcf_record.FORMAT.keys():
                        likelihoods = vcf_record.FORMAT['GL'].split(',')
                        assert(len(likelihoods) > 2)
                        if called_alleles == {'0'}:
                            vcf_record.set_format_key_value('GT_CONF',str(float(likelihoods[0]) - float(likelihoods[1])))
                        else:
                            vcf_record.set_format_key_value('GT_CONF', str(float(likelihoods[int(genotypes[0])]) - float(likelihoods[0])))
                    if 'SupportFraction' in vcf_record.INFO.keys() and 'GT_CONF' not in vcf_record.FORMAT.keys():
                        vcf_record.set_format_key_value('GT_CONF',
                                                        str(float(vcf_record.INFO['SupportFraction'])*100))
                    print(vcf_record, file=f)


    @classmethod
    def _write_vars_plus_flanks_to_fasta(cls, outfile, vcf_records, ref_seqs, flank_length):
        '''Given a dict of vcf records made by vcf_file_read.vcf_file_to_dict(),
        and its correcsponding file of reference sequences, writes a new fasta file
        of each ref seq and inferred variant sequence plus flank_length nucleotides added to
        its start and end. Calls each sequence:
            ref_name.start_position.vcf_list_index.allele_number
        where allele_numbers in same order as VCF, with ref seq = allele 0.'''
        with open(outfile, 'w') as f:
            for ref_name in sorted(vcf_records):
                for i, vcf_record in enumerate(vcf_records[ref_name]):
                    start_position, alleles = vcf_record.inferred_var_seqs_plus_flanks(ref_seqs[ref_name], flank_length)
                    for allele_index, allele_seq in enumerate(alleles):
                        seq_name = '.'.join([ref_name, str(start_position + 1), str(i), str(allele_index)])
                        print('>' + seq_name, allele_seq, sep='\n', file=f)


    @classmethod
    def _map_seqs_to_ref(cls, seqs_file, ref_file, outfile):
        '''Map seqs_file to ref_file using BWA MEM.
        Output is SAM file written to outfile'''
        bwa_binary = dependencies.find_binary('bwa')
        command = ' '.join([
            bwa_binary, 'mem',
            '-a', # report all mappings
            '-Y', # use soft clipping for supplementary alignments
            ref_file,
            seqs_file,
            '>', outfile,
        ])
        utils.syscall(command)


    @classmethod
    def _check_called_genotype(cls, vcf_record):
        if 'GT' in vcf_record.FORMAT:
            called_alts = list(set(vcf_record.FORMAT['GT'].split('/')))
            if len(called_alts) > 1:
                vcf_record.set_format_key_value('MINOS_CHECK_GENOTYPE', 'HET')
            else:
                called_alt = int(called_alts[0])
                pass_per_alt = vcf_record.FORMAT['MINOS_CHECK_ALLELES'].split(',')
                if pass_per_alt[called_alt] == 'Pass':
                    check_geno = '1'
                elif pass_per_alt[called_alt] == 'Exclude':
                    check_geno = 'Exclude'
                else:
                    check_geno = '0'
                vcf_record.set_format_key_value('MINOS_CHECK_GENOTYPE', check_geno)
        else:
            vcf_record.set_format_key_value('MINOS_CHECK_GENOTYPE', 'UNKNOWN_NO_GT')


    @classmethod
    def _check_if_sam_match_is_good(cls, sam_record, ref_seqs, flank_length, query_sequence=None, allow_mismatches=True, max_soft_clipped=3):
        logging.debug(f'Checking SAM: {sam_record}')

        if sam_record.is_unmapped:
            return 'Unmapped'

        if not allow_mismatches:
            try:
                nm = sam_record.get_tag('NM')
            except:
                raise Error('No NM tag found in sam record:' + str(sam_record))

            all_mapped = len(sam_record.cigartuples) == 1 and sam_record.cigartuples[0][0] == 0
            if all_mapped and nm == 0:
                logging.debug('SAM record passed no mismatches allowed check')
                return 'Good'
            else:
                logging.debug('SAM record failed no mismatches allowed check')
                return 'Bad_mismatches'

        # don't allow too many soft clipped bases
        if (sam_record.cigartuples[0][0] == 4 and sam_record.cigartuples[0][1] > max_soft_clipped) \
                or (sam_record.cigartuples[-1][0] == 4 and sam_record.cigartuples[-1][1] > max_soft_clipped):
            logging.debug('SAM record failed soft clipping check')
            return 'Bad_soft_clipped'

        if query_sequence is None:
            query_sequence = sam_record.query_sequence
        assert query_sequence is not None

        assert sam_record.reference_name in ref_seqs

        # if the query is short, which happens when the variant we
        # are checking is too near the start or end of the ref sequence
        if len(query_sequence) < 2 * flank_length + 1:
            # This is an edge case. We don't really know which part
            # of the query seq we're looking for, so guess
            length_diff = 2 * flank_length - len(query_sequence)

            if sam_record.query_alignment_start < 5:
                alt_seq_end = len(query_sequence) - flank_length - 1
                alt_seq_start = min(alt_seq_end, flank_length - length_diff)
            else:
                alt_seq_start = flank_length
                alt_seq_end = max(alt_seq_start, length_diff + len(query_sequence) - flank_length - 1)
        else:
            alt_seq_start = flank_length
            alt_seq_end = len(query_sequence) - flank_length - 1

        aligned_pairs = sam_record.get_aligned_pairs()
        logging.debug(f'aligned_pairs: {aligned_pairs}')
        wanted_aligned_pairs = []
        current_pos = 0

        i = 0
        while i < len(query_sequence):
            if aligned_pairs[i][0] is None:
                if alt_seq_start - 1 <= current_pos <= alt_seq_end + 1:
                    wanted_aligned_pairs.append(aligned_pairs[i])
            elif current_pos > alt_seq_end:
                break
            else:
                current_pos = aligned_pairs[i][0]
                if alt_seq_start - 1 <= current_pos <= alt_seq_end + 1:
                    wanted_aligned_pairs.append(aligned_pairs[i])

            i += 1

        logging.debug(f'wanted_aligned_pairs: {wanted_aligned_pairs}')
        assert len(wanted_aligned_pairs) > 0

        for pair in wanted_aligned_pairs:
            if None in pair or query_sequence[pair[0]] != ref_seqs[sam_record.reference_name][pair[1]]:
                logging.debug('SAM record failed because mismatch in allele sequence plus 1bp either side')
                return 'Bad_allele_mismatch'

        logging.debug('SAM record passed all checks')
        return 'Good'


    @classmethod
    def _parse_sam_file_and_update_vcf_records_and_gather_stats(cls, infile, vcf_records, flank_length, ref_seqs, allow_mismatches=True, exclude_regions=None, max_soft_clipped=3):
        '''Input is SAM file made by _map_seqs_to_ref(), and corresponding dict
        of VCF records made by vcf_file_read.file_to_dict.
        Adds validation info to each VCF record. Returns a dict of stats that
        has counts up the values in MINOS_CHECK_GENOTYPE in the output file:
            total => total number of VCF records
            HET => number of records with a heterozygous call
            'gt_wrong' => number of records where genotype is incorrect
            'gt_correct' => number of records where genotype is correct
            'gt_excluded' => number of records excluded because matched to a position in exclude_regions'''
        if  exclude_regions is None:
            exclude_regions = {}

        stats = {x: 0 for x in ['total', '0', '1', 'HET', 'UNKNOWN_NO_GT', 'Exclude', 'tp_edit_dist', 'fp_edit_dist']}
        gt_conf_hists = {'TP': {}, 'FP': {}}
        sreader = sam_reader(infile)

        for sam_list in sreader:
            stats['total'] += 1

            # Split the sam_list (Which has all records for all alleles from one
            # VCf line) into a list per allele.
            sam_records_by_allele = []
            current_allele_index = None
            vcf_probe_info_tuples = set()
            for sam_record in sam_list:
                vcfref_name, expected_start, vcf_record_index, allele_index = sam_record.query_name.rsplit('.', maxsplit=3)
                vcf_probe_info_tuples.add((vcfref_name, expected_start, vcf_record_index))
                if current_allele_index is None or current_allele_index != allele_index:
                    sam_records_by_allele.append([sam_record])
                else:
                    sam_records_by_allele[-1].append(sam_record)
                current_allele_index = allele_index

            assert len(vcf_probe_info_tuples) == 1
            vcfref_name, expected_start, vcf_record_index = vcf_probe_info_tuples.pop()
            expected_start = int(expected_start) - 1
            vcf_record_index = int(vcf_record_index)
            results = {x: [] for x in ['MINOS_CHECK_ALLELES']}

            for allele_sam_list in sam_records_by_allele:
                # Important! only the first hit actually has the sequence!
                # So give that to MappingBasedVerifier._check_if_sam_match_is_good()
                match_result_types = set()
                for i in range(len(allele_sam_list)):
                    if allele_sam_list[i].is_unmapped:
                        match_result_types.add('Unmapped')
                    else:
                        exclude = False
                        truth_name = allele_sam_list[i].reference_name
                        if truth_name in exclude_regions:
                            start = allele_sam_list[i].reference_start
                            end = allele_sam_list[i].reference_end - 1
                            interval = pyfastaq.intervals.Interval(start, end)
                            exclude = MappingBasedVerifier._interval_intersects_an_interval_in_list(interval, exclude_regions[truth_name])

                        if exclude:
                            match_result_types.add('Exclude')
                        else:
                            match_type = MappingBasedVerifier._check_if_sam_match_is_good(allele_sam_list[i], ref_seqs, flank_length, query_sequence=allele_sam_list[0].query_sequence, allow_mismatches=allow_mismatches, max_soft_clipped=max_soft_clipped)
                            match_result_types.add(match_type)
                            #if good_match:
                            #    match_result_types.add('1')
                            #else:
                            #    match_result_types.add('0')


                #indexes_of_good_matches = [i for i in range(len(allele_sam_list)) if MappingBasedVerifier._check_if_sam_match_is_good(allele_sam_list[i], ref_seqs, flank_length, query_sequence=allele_sam_list[0].query_sequence, allow_mismatches=allow_mismatches)]

                #if len(indexes_of_good_matches) > 0:
                #    results['MINOS_CHECK_ALLELES'].append('1')
                #else:
                #    results['MINOS_CHECK_ALLELES'].append('0')

                logging.debug(f'match_result_types: {match_result_types}')
                if 'Good' in match_result_types:
                    results['MINOS_CHECK_ALLELES'].append('Pass')
                elif 'Exclude' in match_result_types:
                    results['MINOS_CHECK_ALLELES'].append('Exclude')
                else:
                    results['MINOS_CHECK_ALLELES'].append('Fail-' + '-'.join(sorted(list(match_result_types))))

            vcf_record = vcf_records[vcfref_name][vcf_record_index]

            for key in sorted(results):
                vcf_record.set_format_key_value(key, ','.join(results[key]))

            MappingBasedVerifier._check_called_genotype(vcf_record)
            stats[vcf_record.FORMAT['MINOS_CHECK_GENOTYPE']] += 1
            MappingBasedVerifier._add_edit_distances_to_vcf_record(vcf_record)

            if 'GT_CONF' in vcf_record.FORMAT and vcf_record.FORMAT['MINOS_CHECK_GENOTYPE'] in {'0', '1'}:
                tp_or_fp = {'1': 'TP', '0': 'FP'}[vcf_record.FORMAT['MINOS_CHECK_GENOTYPE']]
                gt_conf = int(float(vcf_record.FORMAT['GT_CONF']))
                gt_conf_hists[tp_or_fp][gt_conf] = gt_conf_hists[tp_or_fp].get(gt_conf, 0) + 1

            # if it's a homozygous call, update the edit distance totals
            if 'GT' in vcf_record.FORMAT and len(set(vcf_record.FORMAT['GT'].split('/'))) == 1:
                if vcf_record.FORMAT['MINOS_CHECK_GENOTYPE'] == '1':
                    stats['tp_edit_dist'] += int(float(vcf_record.FORMAT['GT_EDIT_DIST']))
                elif vcf_record.FORMAT['MINOS_CHECK_GENOTYPE'] == '0':
                    stats['fp_edit_dist'] += int(float(vcf_record.FORMAT['GT_EDIT_DIST']))


        stats['gt_wrong'] = stats['0']
        del stats['0']
        stats['gt_correct'] = stats['1']
        del stats['1']
        stats['gt_excluded'] = stats['Exclude']
        del stats['Exclude']
        return stats, gt_conf_hists


    @classmethod
    def _get_missing_vcf_records(cls, called_vcfs, expected_vcfs):
        missing_records = {}

        for ref_seq in expected_vcfs:
            if ref_seq not in called_vcfs:
                missing_records[ref_seq] = expected_vcfs[ref_seq]
                continue

            i = 0
            called_vcf_list = called_vcfs[ref_seq]
            missing_records[ref_seq] = []

            for expected_record in expected_vcfs[ref_seq]:
                while i < len(called_vcf_list) and called_vcf_list[i].ref_end_pos() < expected_record.POS:
                    i += 1

                if i == len(called_vcf_list):
                    missing_records[ref_seq].append(expected_record)
                elif called_vcf_list[i].POS > expected_record.POS or expected_record.ref_end_pos() > called_vcf_list[i].ref_end_pos():
                    missing_records[ref_seq].append(expected_record)

        return missing_records


    @classmethod
    def _get_total_length_of_expected_regions_called(cls, expected_regions, called_vcf_records):
        '''Returns total length of expected regions, and total length that was called'''
        total_length = 0
        called_length = 0

        for ref_seq, expected_list in expected_regions.items():
            if ref_seq in called_vcf_records:
                called_list = [pyfastaq.intervals.Interval(x.POS, x.ref_end_pos()) for x in called_vcf_records[ref_seq]]
                pyfastaq.intervals.merge_overlapping_in_list(called_list)
                total_length += pyfastaq.intervals.length_sum_from_list(expected_list)
                intersect_regions = pyfastaq.intervals.intersection(called_list, expected_list)
                called_length += pyfastaq.intervals.length_sum_from_list(intersect_regions)
            else:
                total_length += pyfastaq.intervals.length_sum_from_list(expected_list)

        return total_length, called_length


    def run(self):
        if self.filter_and_cluster_vcf:
            MappingBasedVerifier._filter_vcf_for_clustering(self.vcf_file_in, self.filtered_vcf, discard_ref_calls=self.discard_ref_calls)
            if self.discard_ref_calls:
                clusterer = vcf_clusterer.VcfClusterer([self.filtered_vcf], self.vcf_reference_file, self.clustered_vcf, merge_method='simple', max_distance_between_variants=self.merge_length)
            else:
                clusterer = vcf_clusterer.VcfClusterer([self.filtered_vcf], self.vcf_reference_file, self.clustered_vcf, merge_method='gt_aware', max_distance_between_variants=self.merge_length)
            clusterer.run()

        vcf_header, vcf_records = vcf_file_read.vcf_file_to_dict(self.vcf_to_check, sort=True, remove_useless_start_nucleotides=True)
        sample_from_header = vcf_file_read.get_sample_name_from_vcf_header_lines(vcf_header)
        if sample_from_header is None:
            sample_from_header = 'sample'
        vcf_ref_seqs = {}
        pyfastaq.tasks.file_to_dict(self.vcf_reference_file, vcf_ref_seqs)
        verify_ref_seqs = {}
        pyfastaq.tasks.file_to_dict(self.verify_reference_file, verify_ref_seqs)

        MappingBasedVerifier._write_vars_plus_flanks_to_fasta(self.seqs_out, vcf_records, vcf_ref_seqs, self.flank_length)
        MappingBasedVerifier._map_seqs_to_ref(self.seqs_out, self.verify_reference_file, self.sam_file_out)
        os.unlink(self.seqs_out)
        stats, gt_conf_hists = MappingBasedVerifier._parse_sam_file_and_update_vcf_records_and_gather_stats(self.sam_file_out, vcf_records, self.flank_length, verify_ref_seqs, allow_mismatches=self.allow_flank_mismatches, exclude_regions=self.exclude_regions, max_soft_clipped=self.max_soft_clipped)

        with open(self.vcf_file_out, 'w') as f:
            print(*vcf_header, sep='\n', file=f)
            for r in vcf_records:
                for v in vcf_records[r]:
                    print(v, file=f)

        # false negative stats, if possible
        stats['variant_regions_total'] = 'NA'
        stats['called_variant_regions'] = 'NA'

        if self.run_dnadiff:
            dnadiffer = dnadiff.Dnadiff(
                self.verify_reference_file,
                self.vcf_reference_file,
                self.dnadiff_outprefix,
            )
            dnadiffer.run()
            stats['variant_regions_total'], stats['called_variant_regions'] = MappingBasedVerifier._get_total_length_of_expected_regions_called(dnadiffer.all_variant_intervals, vcf_records)
            expected_variants = dnadiffer.variants
        elif self.expected_variants_vcf is not None:
            header, expected_variants = vcf_file_read.vcf_file_to_dict(self.expected_variants_vcf, sort=True, remove_useless_start_nucleotides=True)
        else:
            expected_variants = None

        if expected_variants is None:
            stats['false_negatives'] = 'NA'
        else:
            missed_vcf_records = MappingBasedVerifier._get_missing_vcf_records(vcf_records, expected_variants)
            stats['false_negatives'] = 0
            with open(self.vcf_false_negatives_file_out, 'w') as f:
                print('##fileformat=VCFv4.2', file=f)
                print('#CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'INFO', 'FORMAT', sample_from_header, sep='\t', file=f)
                for vcf_list in missed_vcf_records.values():
                    stats['false_negatives'] += len(vcf_list)
                    if len(vcf_list) > 0:
                        print(*vcf_list, sep='\n', file=f)

        # write stats file
        with open(self.stats_out, 'w') as f:
            keys = ['total', 'gt_correct', 'gt_wrong', 'gt_excluded', 'HET', 'tp_edit_dist', 'fp_edit_dist', 'UNKNOWN_NO_GT', 'variant_regions_total', 'called_variant_regions', 'false_negatives']
            print(*keys, sep='\t', file=f)
            print(*[stats[x] for x in keys], sep='\t', file=f)


        # write GT_CONG histogram files
        for key, filename in self.gt_conf_hists_filenames.items():
            with open(filename, 'w') as f:
                print('GT_CONF\tCount', file=f)
                for gt_conf, count in sorted(gt_conf_hists[key].items()):
                    print(gt_conf, count, sep='\t', file=f)

        plots.plots_from_minos_vcf(self.vcf_file_out, self.vcf_file_plots_out)




from minos import vcf_chunker

def run(options):
    chunker = vcf_chunker.VcfChunker(
        options.outdir,
        vcf_infile=options.vcf_file,
        ref_fasta=options.ref_fasta,
        variants_per_split=options.variants_per_split,
        alleles_per_split=options.alleles_per_split,
        total_splits=options.total_splits,
        max_read_length=options.max_read_length,
        flank_length=options.max_read_length,
        gramtools_kmer_size=options.gramtools_kmer_size,
        threads=options.threads,
    )
    chunker.make_split_files()





import json
import logging
import os
import shutil
import statistics
import sys

from cluster_vcf_records import vcf_clusterer, vcf_file_read

from minos import bam_read_extract, dependencies, genotype_confidence_simulator, gramtools, plots, utils, vcf_chunker

class Error (Exception): pass

class Adjudicator:
    def __init__(self,
        outdir,
        ref_fasta,
        reads_files,
        vcf_files,
        max_read_length=None,
        read_error_rate=None,
        overwrite_outdir=False,
        max_alleles_per_cluster=5000,
        gramtools_build_dir=None,
        gramtools_kmer_size=10,
        sample_name=None,
        variants_per_split=None,
        alleles_per_split=None,
        total_splits=None,
        clean=True,
        genotype_simulation_iterations=10000,
        use_unmapped_reads=False,
    ):
        self.ref_fasta = os.path.abspath(ref_fasta)
        self.reads_files = [os.path.abspath(x) for x in reads_files]
        self.vcf_files = [os.path.abspath(x) for x in vcf_files]
        self.overwrite_outdir = overwrite_outdir
        self.max_alleles_per_cluster = max_alleles_per_cluster
        self.sample_name = sample_name
        self.outdir = os.path.abspath(outdir)
        self.split_output_dir = os.path.join(self.outdir, 'split.out')
        self.log_file = os.path.join(self.outdir, 'log.txt')
        self.clustered_vcf = os.path.join(self.outdir, 'gramtools.in.vcf')
        self.unfiltered_vcf_file = os.path.join(self.outdir, 'debug.calls_with_zero_cov_alleles.vcf')
        self.final_vcf = os.path.join(self.outdir, 'final.vcf')
        self.plots_prefix = os.path.join(self.outdir, 'final.vcf.plots')

        if gramtools_build_dir is None:
            self.split_input_dir = os.path.join(self.outdir, 'split.in')
            self.gramtools_build_dir = os.path.join(self.outdir, 'gramtools.build')
            self.user_supplied_gramtools_build_dir = False
        else:
            self.gramtools_build_dir = os.path.abspath(gramtools_build_dir)
            self.split_input_dir = self.gramtools_build_dir
            self.user_supplied_gramtools_build_dir = True
            if not os.path.exists(self.gramtools_build_dir):
                raise Error('Error! gramtools_build_dir=' + self.gramtools_build_dir + ' used, but directory not found on disk. Cannot continue')
            if len(self.vcf_files) != 1:
                raise Error('Error! If gramtools_build_dir is used, then There Can Be Only One input VCF file (which is assumed to be clustered')


        self.gramtools_kmer_size = gramtools_kmer_size
        self.gramtools_quasimap_dir = os.path.join(self.outdir, 'gramtools.quasimap')
        self.perl_generated_vcf = os.path.join(self.gramtools_build_dir, 'perl_generated_vcf')
        self.read_error_rate = read_error_rate
        self.max_read_length = max_read_length
        self.variants_per_split = variants_per_split
        self.alleles_per_split = alleles_per_split
        self.total_splits = total_splits

        if (self.total_splits is not None or self.variants_per_split is not None or self.alleles_per_split is not None) and len(self.reads_files) != 1:
            raise Error('Error! If using splitting, must input one reads file (which is assumed to be a sorted indexed BAM file)')

        self.clean = clean
        self.genotype_simulation_iterations = genotype_simulation_iterations
        self.use_unmapped_reads = use_unmapped_reads


    @classmethod
    def _get_gramtools_kmer_size(cls, build_dir, input_kmer_size):
        if build_dir is not None and os.path.exists(build_dir):
            json_file = os.path.join(build_dir, 'build_report.json')
            with open(json_file) as f:
                json_build_report = json.load(f)

            if 'kmer_size' in json_build_report:
                kmer_size_from_json = json_build_report['kmer_size']
            else:
                raise Error('kmer_size not found in graamtools build report' + json_file)

            if input_kmer_size is not None:
                logging.warning('gramtools_kmer_size option used, but going to ignore it because gramtools_build_dir used, which means that gramtools build has already been run')

            return kmer_size_from_json
        elif input_kmer_size is None:
            return 10
        else:
            return input_kmer_size


    def run(self):
        if os.path.exists(self.outdir) and self.overwrite_outdir:
            shutil.rmtree(self.outdir)

        try:
            os.mkdir(self.outdir)
        except:
            raise Error('Error making output directory ' + self.outdir)

        fh = logging.FileHandler(self.log_file, mode='w')
        log = logging.getLogger()
        formatter = logging.Formatter('[minos %(asctime)s %(levelname)s] %(message)s', datefmt='%d-%m-%Y %H:%M:%S')
        fh.setFormatter(formatter)
        log.addHandler(fh)
        logging.info('Command run: ' + ' '.join(sys.argv))
        dependencies.check_and_report_dependencies(programs=['gramtools'])
        logging.info('Dependencies look OK')

        if self.read_error_rate is None or self.max_read_length is None:
            logging.info('One or both of read_error_rate and max_read_length not known. Estimate from first 10,000 reads...')
            estimated_read_length, estimated_read_error_rate = utils.estimate_max_read_length_and_read_error_rate_from_qual_scores(self.reads_files[0])
            logging.info('Estimated max_read_length=' + str(estimated_read_length) + ' and read_error_rate=' + str(estimated_read_error_rate))

        self.read_error_rate = estimated_read_error_rate if self.read_error_rate is None else self.read_error_rate
        self.max_read_length = estimated_read_length if self.max_read_length is None else self.max_read_length
        logging.info('Using max_read_length=' + str(self.max_read_length) + ' and read_error_rate=' + str(self.read_error_rate))

        if self.user_supplied_gramtools_build_dir:
            logging.info('User supplied gramtools build dir. Assuming VCF already clustered, so skipping clustering')
            assert len(self.vcf_files) == 1
            self.clustered_vcf = self.vcf_files[0]
        else:
            logging.info('Clustering VCF file(s), to make one VCF input file for gramtools')
            clusterer = vcf_clusterer.VcfClusterer(
                self.vcf_files,
                self.ref_fasta,
                self.clustered_vcf,
                max_distance_between_variants=1,
                max_alleles_per_cluster=self.max_alleles_per_cluster,
            )
            clusterer.run()

            logging.info('Finished clustering VCF file(s)')

        if not vcf_file_read.vcf_file_has_at_least_one_record(self.clustered_vcf):
            error_message = 'No VCF records. Cannot continue. Please check that the input VCF files contained at least one variant'
            logging.error(error_message)
            raise Error(error_message)


        if self.total_splits is not None or self.variants_per_split is not None or self.alleles_per_split is not None or os.path.exists(os.path.join(self.split_input_dir, 'data.pickle')):
            self._run_gramtools_with_split_vcf()
        else:
            self._run_gramtools_not_split_vcf()

        logging.info('Making plots from final.vcf')
        plots.plots_from_minos_vcf(self.final_vcf, self.plots_prefix)

        logging.info('All done! Thank you for using minos :)')


    @classmethod
    def _add_gt_conf_percentile_and_filters_to_vcf_file(cls, vcf_file, mean_depth, depth_variance, error_rate, iterations, min_dp=2, min_gt_conf_percentile=2.5):
        '''Overwrites vcf_file, with new version that has GT_CONF_PERCENTILE added,
        and filter for DP and GT_CONF_PERCENTILE'''
        simulations = genotype_confidence_simulator.GenotypeConfidenceSimulator(mean_depth, depth_variance, error_rate, allele_length=1, iterations=iterations)
        simulations.run_simulations()
        vcf_header, vcf_lines = vcf_file_read.vcf_file_to_list(vcf_file)
        for i, line in enumerate(vcf_header):
            if line.startswith('##FORMAT=<ID=GT_CONF'):
                break
        else:
            raise Exception(f'No GT_CONF description found in header of VCF file {vcf_file}. Cannot continue')

        vcf_header.insert(i+1, r'''##FORMAT=<ID=GT_CONF_PERCENTILE,Number=1,Type=Float,Description="Percentile of GT_CONF">''')
        vcf_header.insert(i+1, f'##FILTER=<ID=MIN_DP,Description="Minimum DP of {min_dp}">')
        vcf_header.insert(i+1, f'##FILTER=<ID=MIN_GCP,Description="Minimum GT_CONF_PERCENTILE of {min_gt_conf_percentile}">')

        with open(vcf_file, 'w') as f:
            print(*vcf_header, sep='\n', file=f)

            for vcf_record in vcf_lines:
                vcf_record.FILTER = set()

                if 'GT_CONF' in vcf_record.FORMAT:
                    conf = int(round(float(vcf_record.FORMAT['GT_CONF'])))
                    if 'GT' in vcf_record.FORMAT and '.' not in vcf_record.FORMAT['GT']:
                        vcf_record.set_format_key_value('GT_CONF_PERCENTILE', str(simulations.get_percentile(conf)))
                        if 'DP' in vcf_record.FORMAT and float(vcf_record.FORMAT['DP']) < min_dp:
                            vcf_record.FILTER.add('MIN_DP')
                        if float(vcf_record.FORMAT['GT_CONF_PERCENTILE']) < min_gt_conf_percentile:
                            vcf_record.FILTER.add('MIN_GCP')
                        if len(vcf_record.FILTER) == 0:
                            vcf_record.FILTER.add('PASS')

                print(vcf_record, file=f)


    def _run_gramtools_not_split_vcf(self):
        self.gramtools_kmer_size = Adjudicator._get_gramtools_kmer_size(self.gramtools_build_dir, self.gramtools_kmer_size)
        build_report, quasimap_report = gramtools.run_gramtools(
            self.gramtools_build_dir,
            self.gramtools_quasimap_dir,
            self.clustered_vcf,
            self.ref_fasta,
            self.reads_files,
            self.max_read_length,
            kmer_size=self.gramtools_kmer_size,
        )

        logging.info('Loading gramtools quasimap output files ' + self.gramtools_quasimap_dir)
        mean_depth, depth_variance, vcf_header, vcf_records, allele_coverage, allele_groups = gramtools.load_gramtools_vcf_and_allele_coverage_files(self.perl_generated_vcf, self.gramtools_quasimap_dir)
        logging.info('Finished loading gramtools files')
        if self.sample_name is None:
            sample_name = vcf_file_read.get_sample_name_from_vcf_header_lines(vcf_header)
        else:
            sample_name = self.sample_name
        assert sample_name is not None
        logging.info('Writing VCf output file ' + self.final_vcf)
        gramtools.write_vcf_annotated_using_coverage_from_gramtools(
            mean_depth,
            vcf_records,
            allele_coverage,
            allele_groups,
            self.read_error_rate,
            self.unfiltered_vcf_file,
            self.gramtools_kmer_size,
            sample_name=sample_name,
            max_read_length=self.max_read_length,
            filtered_outfile=self.final_vcf
        )

        logging.info(f'Adding GT_CONF_PERCENTLE to debug VCF file {self.unfiltered_vcf_file}, using mean depth {mean_depth}, depth variance {depth_variance}, error rate {self.read_error_rate}, and {self.genotype_simulation_iterations} simulation iterations')
        Adjudicator._add_gt_conf_percentile_and_filters_to_vcf_file(self.unfiltered_vcf_file, mean_depth, depth_variance, self.read_error_rate, self.genotype_simulation_iterations)
        logging.info(f'Adding GT_CONF_PERCENTLE to final VCF file {self.final_vcf}, using mean depth {mean_depth}, depth variance {depth_variance}, error rate {self.read_error_rate}, and {self.genotype_simulation_iterations} simulation iterations')
        Adjudicator._add_gt_conf_percentile_and_filters_to_vcf_file(self.final_vcf, mean_depth, depth_variance, self.read_error_rate, self.genotype_simulation_iterations)

        if self.clean:
            os.rename(os.path.join(self.gramtools_quasimap_dir, 'report.json'), os.path.join(self.outdir, 'gramtools.quasimap.report.json'))
            shutil.rmtree(self.gramtools_quasimap_dir)

            if not self.user_supplied_gramtools_build_dir:
                os.rename(os.path.join(self.gramtools_build_dir, 'build_report.json'), os.path.join(self.outdir, 'gramtools.build.report.json'))
                shutil.rmtree(self.gramtools_build_dir)


    def _run_gramtools_with_split_vcf(self):
        logging.info('Splitting VCF files into chunks (if not already done)')
        chunker = vcf_chunker.VcfChunker(
            self.split_input_dir,
            vcf_infile=self.clustered_vcf,
            ref_fasta=self.ref_fasta,
            variants_per_split=self.variants_per_split,
            alleles_per_split=self.alleles_per_split,
            max_read_length=self.max_read_length,
            total_splits=self.total_splits,
            flank_length=self.max_read_length,
            gramtools_kmer_size=self.gramtools_kmer_size,
        )
        chunker.make_split_files()
        self.gramtools_kmer_size = chunker.gramtools_kmer_size

        logging.info('VCF file split into ' + str(chunker.total_split_files) + ' chunks')
        try:
            os.mkdir(self.split_output_dir)
        except:
            raise Error('Error making output split directory ' + self.split_output_dir)

        if self.use_unmapped_reads:
            unmapped_reads_file = os.path.join(self.split_output_dir, 'unmapped_reads.bam')
            bam_read_extract.get_unmapped_reads(self.reads_files[0], unmapped_reads_file)

        split_vcf_outfiles = {}
        split_vcf_outfiles_unfiltered = {}
        mean_depths = []
        depth_variances = []

        for ref_name, split_file_list in chunker.vcf_split_files.items():
            split_vcf_outfiles[ref_name] = []
            split_vcf_outfiles_unfiltered[ref_name] = []
            for split_file in split_file_list:
                logging.info('===== Start analysing variants in VCF split file ' + split_file.filename + ' =====')
                split_reads_file = os.path.join(self.split_output_dir, 'split.' + str(split_file.file_number) + '.reads.bam')
                bam_read_extract.get_region(
                    self.reads_files[0],
                    split_file.chrom,
                    split_file.chrom_start,
                    split_file.chrom_end,
                    split_reads_file,
                )

                gramtools_quasimap_dir = os.path.join(self.split_output_dir, 'split.' + str(split_file.file_number) + '.gramtools.quasimap')
                if self.use_unmapped_reads:
                    reads_files = [unmapped_reads_file, split_reads_file]
                else:
                    reads_files = [split_reads_file]

                build_report, quasimap_report = gramtools.run_gramtools(
                    split_file.gramtools_build_dir,
                    gramtools_quasimap_dir,
                    split_file.filename,
                    self.ref_fasta,
                    reads_files,
                    self.max_read_length,
                    kmer_size=self.gramtools_kmer_size,
                )

                logging.info('Loading split gramtools quasimap output files ' + gramtools_quasimap_dir)
                perl_generated_vcf = os.path.join(split_file.gramtools_build_dir, 'perl_generated_vcf')
                mean_depth, depth_variance, vcf_header, vcf_records, allele_coverage, allele_groups = gramtools.load_gramtools_vcf_and_allele_coverage_files(perl_generated_vcf, gramtools_quasimap_dir)
                mean_depths.append(mean_depth)
                depth_variances.append(depth_variance)
                logging.info('Finished loading gramtools files')
                if self.sample_name is None:
                    sample_name = vcf_file_read.get_sample_name_from_vcf_header_lines(vcf_header)
                else:
                    sample_name = self.sample_name
                assert sample_name is not None
                split_vcf_out = os.path.join(self.split_output_dir, 'split.' + str(split_file.file_number) + '.out.vcf')
                unfiltered_vcf_out = os.path.join(self.split_output_dir, 'split.' + str(split_file.file_number) + '.out.debug.calls_with_zero_cov_alleles.vcf')
                logging.info('Writing VCf output file ' + split_vcf_out + ' for split VCF file ' + split_file.filename)
                gramtools.write_vcf_annotated_using_coverage_from_gramtools(
                    mean_depth,
                    vcf_records,
                    allele_coverage,
                    allele_groups,
                    self.read_error_rate,
                    unfiltered_vcf_out,
                    self.gramtools_kmer_size,
                    sample_name=sample_name,
                    max_read_length=self.max_read_length,
                    filtered_outfile=split_vcf_out,
                )
                split_vcf_outfiles[ref_name].append(split_vcf_out)
                split_vcf_outfiles_unfiltered[ref_name].append(unfiltered_vcf_out)

                if self.clean:
                    logging.info('Cleaning gramtools files from split VCF file ' + split_file.filename)
                    if not self.user_supplied_gramtools_build_dir:
                        os.rename(os.path.join(split_file.gramtools_build_dir, 'build_report.json'), split_file.gramtools_build_dir + '.report.json')
                        shutil.rmtree(split_file.gramtools_build_dir)
                        os.unlink(split_file.filename)

                    os.rename(os.path.join(gramtools_quasimap_dir, 'report.json'), gramtools_quasimap_dir + '.report.json')
                    shutil.rmtree(gramtools_quasimap_dir)
                    os.unlink(split_reads_file)

                logging.info('===== Finish analysing variants in VCF split file ' + split_file.filename + ' =====')

        logging.info('Merging VCF files into one output file ' + self.final_vcf)
        chunker.merge_files(split_vcf_outfiles, self.final_vcf)
        chunker.merge_files(split_vcf_outfiles_unfiltered, self.unfiltered_vcf_file)

        mean_depth = statistics.mean(mean_depths)
        depth_variance = statistics.mean(depth_variances)
        logging.info(f'Adding GT_CONF_PERCENTLE to debug VCF file {self.unfiltered_vcf_file}, using mean depth {mean_depth}, depth variance {depth_variance}, error rate {self.read_error_rate}, and {self.genotype_simulation_iterations} simulation iterations')
        Adjudicator._add_gt_conf_percentile_and_filters_to_vcf_file(self.unfiltered_vcf_file, mean_depth, depth_variance, self.read_error_rate, self.genotype_simulation_iterations)
        logging.info(f'Adding GT_CONF_PERCENTLE to final VCF file {self.final_vcf}, using mean depth {mean_depth}, depth variance {depth_variance}, error rate {self.read_error_rate}, and {self.genotype_simulation_iterations} simulation iterations')
        Adjudicator._add_gt_conf_percentile_and_filters_to_vcf_file(self.final_vcf, mean_depth, depth_variance, self.read_error_rate, self.genotype_simulation_iterations)

        if self.clean:
            logging.info('Deleting temp split VCF files')
            for d in split_vcf_outfiles, split_vcf_outfiles_unfiltered:
                for file_list in d.values():
                    for filename in file_list:
                        os.unlink(filename)
            if self.use_unmapped_reads:
                os.unlink(unmapped_reads_file)




import subprocess
import sys

import pyfastaq
import pysam

class Error (Exception): pass

def syscall(command, allow_fail=False):
    completed_process = subprocess.run(command, shell=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE, universal_newlines=True)
    if (not allow_fail) and completed_process.returncode != 0:
        print('Error running this command:', command, file=sys.stderr)
        print('Return code:', completed_process.returncode, file=sys.stderr)
        print('\nOutput from stdout:', completed_process.stdout, sep='\n', file=sys.stderr)
        print('\nOutput from stderr:', completed_process.stderr, sep='\n', file=sys.stderr)
        raise Error('Error in system call. Cannot continue')

    return completed_process


def estimate_max_read_length_and_read_error_rate_from_qual_scores(infile, number_of_reads=10000):
    '''Estimates the maximum read length, and error rate from a file of reads, using the
    quality scores. Calculated by converting the mean phred quality score
    into the probability (formula is P = 10 ^ (- mean qual / 10)).
    Returns a tuple max_read_length, error_rate.
    Error rate is None if no quality
    scores found. Uses the first number_of_reads found in the file.
    File type can be BAM, SAM, FASTQ and is auto detected'''
    try:
        f = pysam.AlignmentFile(infile)
        filetype = 'pysam'
    except ValueError:
        f = pyfastaq.sequences.file_reader(infile)
        filetype = 'fastaq'
    except:
        raise Error('Error opening file ' + infile)

    total = 0
    base_count = 0
    read_count = 0
    max_read_length = 0

    for sequence in f:
        read_count += 1
        if read_count > number_of_reads:
            break

        if filetype == 'pysam':
            max_read_length = max(max_read_length, sequence.query_length)
            if sequence.query_qualities is not None:
                for qual_score in sequence.query_qualities:
                    total += qual_score
                base_count += len(sequence.query_qualities)
        else:
            max_read_length = max(max_read_length, len(sequence))
            if type(sequence) != pyfastaq.sequences.Fastq:
                continue

            for qual_score in [ord(x) - 33 for x in sequence.qual]:
                total += qual_score
            base_count += len(sequence)

    if base_count == 0:
        return max_read_length, None

    mean_qual_score = total / base_count
    return max_read_length, pow(10, -mean_qual_score / 10)

